#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Oct 25 18:34:18 2021

@author: sandro
"""

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.distributions import (Dirichlet, Categorical)
import math
import numpy as np
from time import time
from matplotlib import pyplot as plt

class MultivariateNormal(object):
    def __init__(self, dim):
        self.dim = dim

    def sample(self, B, K, labels):
        raise NotImplementedError

    def log_prob(self, X, params):
        raise NotImplementedError

    def stats(self):
        raise NotImplementedError

    def parse(self, raw):
        raise NotImplementedError

class MixtureOfMVNs(object):
    def __init__(self, mvn):
        self.mvn = mvn
        self.device = "cuda" if torch.cuda.is_available() else "cpu"

    def sample(self, B, N, K, return_gt=False):
        pi = Dirichlet(torch.ones(K)).sample(torch.Size([B])).to(self.device)
        labels = Categorical(probs=pi).sample(torch.Size([N])).to(self.device)
        labels = labels.transpose(0,1).contiguous()

        X, params = self.mvn.sample(B, K, labels)
        if return_gt:
            return X, labels, pi, params
        else:
            return X

    def log_prob(self, X, pi, params, return_labels=False):
        ll = self.mvn.log_prob(X, params)
        ll = ll + (pi + 1e-10).log().unsqueeze(-2)
        if return_labels:
            labels = ll.argmax(-1)
            return ll.logsumexp(-1).mean(), labels
        else:
            return ll.logsumexp(-1).mean()

    def parse(self, raw):
        return self.mvn.parse(raw)
    
class MultivariateNormalDiag(MultivariateNormal):
    def __init__(self, dim, sigma=0.3):
        super(MultivariateNormalDiag, self).__init__(dim)
        self.device = "cuda" if torch.cuda.is_available() else "cpu"
        self.sigma = sigma

    def sample(self, B, K, labels):
        N = labels.shape[-1]
        mu = -3 + 6*torch.rand(B, K, self.dim).to(self.device)
        sigma = self.sigma*(0.2 + 0.8*torch.rand(B, K, self.dim).to(self.device))
        eps = torch.randn(B, N, self.dim).to(self.device)

        rlabels = labels.unsqueeze(-1).repeat(1, 1, self.dim)
        X = torch.gather(mu, 1, rlabels) + \
                eps * torch.gather(sigma, 1, rlabels)
        return X, (mu, sigma)

    def log_prob(self, X, params):
        mu, sigma = params
        X = X.unsqueeze(2)
        mu = mu.unsqueeze(1)
        sigma = sigma.unsqueeze(1)
        diff = X - mu
        ll = -0.5*math.log(2*math.pi) - sigma.log() - 0.5*(diff.pow(2)/sigma.pow(2))
        return ll.sum(-1)

    def stats(self, params):
        mu, sigma = params
        I = torch.eye(self.dim)[(None,)*(len(sigma.shape)-1)].to(sigma.device)
        cov = sigma.pow(2).unsqueeze(-1) * I
        return mu, cov

    def parse(self, raw):
        pi = torch.softmax(raw[...,0], -1)
        mu = raw[...,1:1+self.dim]
        sigma = F.softplus(raw[...,1+self.dim:])
        return pi, (mu, sigma)