#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Feb  4 12:01:15 2022

@author: sandro
"""

import random
import string
import yaml
import pickle
import os
import shutil
import torch
from torch.utils.data import Dataset
from torch.utils import tensorboard
import torch.optim as optim
import numpy as np
import time

if __name__ == '__main__':
    from Energy_Tools import hamiltonian
    from util import create_folder, recreate_folder, pbc_diff, pbc_rms_error,\
        pbc_mean_relative_energy_error
else:
    from .Energy_Tools import hamiltonian
    from .util import create_folder, recreate_folder, pbc_diff, pbc_rms_error,\
        pbc_mean_relative_energy_error
        
def apply_PBC_to_distances(distances, box_size=6):
    # Only apply to postion columns
    distances[:,:,-4:-2][distances[:,:,-4:-2] > box_size/2] -= box_size
    distances[:,:,-4:-2][distances[:,:,-4:-2] <= -box_size/2] += box_size
    return distances

# Custom MSE loss that takes periodic boundry conditions into account
def PBC_MSE_loss(output, target, box_size=6):
    # Get difference
    error = output - target
    # Deal with periodic boundry conditions
    error = apply_PBC_to_distances(error, box_size=box_size)
    # Get MSE
    loss = torch.mean((error)**2)
    return loss

class TrajectoryDataset(Dataset):
    def __init__(self, folder_path, split='train', rollout=False, target_step=1):
        
        self.folder_path = folder_path
        self.split = split
        self.split_folder = os.path.join(folder_path, split)

        # Load parameters from yaml file
        with open(os.path.join(self.folder_path, "params.yaml")) as yaml_file:
            params = yaml.load(yaml_file, Loader=yaml.FullLoader)
            self.box_size = float(params["box_size"])
            self.time_step = float(params["dt"])
            self.n_particles = int(params["n_particles"])
            self.trajectory_count = int(params[split+"_size"])
            self.trajectory_len = int(params["timesteps"])+1
            self.softening = bool(params["softening"])
            self.softening_radius = float(params["softening_radius"])
            self.simulation_type = params.get("simulation_type", "gravity")
            self.physical_const = float(params["physical_const"])

        # Set number of features each particle has
        if self.simulation_type == "coulomb":
            self.n_features = 6
        else:
            self.n_features = 5

        # Set how many steps into the future target will be. By default target is the next step
        self.target_step = target_step
        
        # Create a numpy array to store all trajectories
        if self.n_particles < 10000 or self.split == 'test':
            self.trajectories = torch.zeros((self.trajectory_count, self.trajectory_len,
                                             self.n_particles, self.n_features)) 
                                            # each particle state is [m, x, y, v_x, v_y]
        else:
            self.trajectories = None

        # If rollout data is requested return full trajectories,
        # otherwise one t, t+1 pair is returned
        self.rollout = rollout

        # Count total number of training samples available
        if self.rollout:
            self.no_of_samples = self.trajectory_count
        else:
            self.no_of_samples = (self.trajectory_len - self.target_step) *\
                                                            self.trajectory_count

        # Populate only the trajectory tensor if it exists/fits into memory
        if self.trajectories is not None:
            for i in range(self.trajectory_count):
                trajectory_path = os.path.join(self.split_folder,
                                               f"simulated_trajectory_{i}.npy")
                trajectory = np.load(trajectory_path)
                self.trajectories[i] = torch.from_numpy(trajectory.astype(np.float32))
        else:
            # Pre-parse trajectories into tensors for each time step in trajectory
            # if not done already (not needed for for test/rollout)
            self.trajectory_tensor_folder = os.path.join(self.split_folder, 'tensors')
            if not os.path.isdir(self.trajectory_tensor_folder) and not self.rollout:
                create_folder(self.trajectory_tensor_folder)
                for i in range(self.trajectory_count):
                    trajectory_path = os.path.join(self.split_folder,
                                                   f"simulated_trajectory_{i}.npy")
                    trajectory = np.load(trajectory_path)
                    for t, tensor in enumerate(trajectory):
                        tensor_path = os.path.join(self.trajectory_tensor_folder,
                                                   f"simulated_trajectory_{i}_{t}.tpkl")
                        torch.save(torch.from_numpy(tensor.astype(np.float32)),
                                   tensor_path)

    def __len__(self):
        
        return self.no_of_samples

    def __getitem__(self, idx):
        
        if self.rollout:
            # Dynamically load the trajectories if they are not in memory
            if self.trajectories is not None:
                trajectory = self.trajectories[idx,::self.target_step]
            else:
                trajectory = torch.from_numpy(np.load(os.path.join(self.split_folder,
                    f"simulated_trajectory_{idx}.npy")).astype(np.float32))[::self.target_step]
            # Return: inputs, targets  
            return trajectory[:-1], trajectory[1:]
        else:
            # Get trajectory from idx
            trajectory_idx = idx // (self.trajectory_len - self.target_step)
            # Get particular time step from idx
            t_idx = idx % (self.trajectory_len - self.target_step)
            if self.trajectories is not None:
                trajectory = self.trajectories[trajectory_idx]
                trajectory_input, trajectory_target = trajectory[t_idx],\
                                            trajectory[t_idx + self.target_step]
            else:
                trajectory_input = torch.load(os.path.join(self.trajectory_tensor_folder,
                                    f"simulated_trajectory_{trajectory_idx}_{t_idx}.tpkl"))
                trajectory_target = torch.load(os.path.join(self.trajectory_tensor_folder,
                    f"simulated_trajectory_{trajectory_idx}_{t_idx + self.target_step}.tpkl"))
            # Return: inputs, targets  
            return trajectory_input, trajectory_target
        
def train_model(model, dataset="3_particles_gravity", learning_rate=1e-3,
                lr_decay=0.97725, batch_size=100, epochs=200, accumulate_steps=1,
                model_dir="models", data_dir="Particle_Movemet/data",
                validate=True, validate_epochs=1,
                data_loader_workers=2, smooth_lr_decay=False,
                target_step=1, cpu=False, experiment_dir="", log_dir="runs",
                resume_checkpoint="", save_after_time=0):
    
    # Track time for saving after x seconds
    start_time_for_save = time.monotonic()

    # Set CPU paralelization to maximum
    torch.set_num_threads(torch.get_num_threads())

    # Load training dataset
    train_set = TrajectoryDataset(folder_path=os.path.join(data_dir, dataset),
                                  split='train', target_step=target_step)
    train_loader = torch.utils.data.DataLoader(train_set, batch_size=batch_size,
                        shuffle=True, num_workers=data_loader_workers, pin_memory=True)

    # Load validation dataset
    validation_set = TrajectoryDataset(folder_path=os.path.join(data_dir, dataset),
                                       split='validation', target_step=target_step)
    validation_loader = torch.utils.data.DataLoader(validation_set, batch_size=batch_size,
                        shuffle=False, num_workers=data_loader_workers, pin_memory=True)

    # Get parameters form dataset
    box_size = train_set.box_size
    time_step = train_set.time_step
    n_particles = train_set.n_particles
    simulation_type = train_set.simulation_type

    optimizer = optim.Adam(model.parameters(), lr=learning_rate)#3e-4
    # decay every 2 * 10^5 with lower imit of 10^-7
    scheduler = optim.lr_scheduler.ExponentialLR(optimizer, lr_decay) 

    # Track iterations and running loss for loging
    n_iter = 0
    running_loss = 0.0

    device = torch.device("cuda:0" if ((not cpu) and torch.cuda.is_available()) else "cpu")

    # Use current/start time to identify saved model and log dir
    start_time = time.strftime("%Y%m%d-%H%M%S")

    if len(resume_checkpoint) > 0:
        print('Resuming checkpoint')
        model_name = resume_checkpoint.replace('_checkpoint.tar', '')
        checkpoint = torch.load(os.path.join(model_dir, experiment_dir, dataset,
                                             resume_checkpoint))
        model.load_state_dict(checkpoint['model_state_dict'])
        optimizer.load_state_dict(checkpoint['optimizer_state_dict'])
        scheduler.load_state_dict(checkpoint['scheduler_state_dict'])
        starting_epoch = checkpoint['epoch']
        n_iter = starting_epoch * len(train_loader)
    else:
        # Create model filename to use for saved model params and tensorboard log dir
        rand_string = ''.join(random.choices(string.ascii_uppercase + string.digits, k=10))
        model_name = "model" +\
            f"_lr_{learning_rate}_decay_{lr_decay}_epochs_{epochs}_batch" +\
            f"_size_{batch_size}_accumulate_steps_{accumulate_steps}" +\
            f"_target_step_{target_step}_{start_time}_{rand_string}"
        starting_epoch = 0

    # Setup direcotries for logs and models
    if len(experiment_dir) > 0:
        logdir=os.path.join(log_dir, experiment_dir, dataset, model_name)
        model_save_path = os.path.join(model_dir, experiment_dir, dataset)
    else:
        logdir=os.path.join(log_dir, dataset, model_name)
        model_save_path = os.path.join(model_dir, dataset)
    
    # Setup Writer for Tensorboard
    writer = tensorboard.SummaryWriter(log_dir=logdir)

    # Log lr for tensorboard
    writer.add_scalar('LearningRate/optimizer_lr', optimizer.param_groups[0]['lr'], n_iter)

    # Create output folder for the dataset models
    create_folder(model_save_path)

    # Track forward pass time
    forward_pass_times = []

    print("Training model %s on %s dataset" % (model_name, dataset))

    for epoch in range(starting_epoch, epochs):
        optimizer.zero_grad()
        for i, data in enumerate(train_loader, 0):
            # Do one training step and get loss value
            inputs, targets = data
            inputs = inputs.to(device, non_blocking=True)
            targets = targets.to(device, non_blocking=True)
            start_time = time.perf_counter_ns()
            outputs = model(inputs)
            end_time = time.perf_counter_ns()
            loss = PBC_MSE_loss(outputs, targets[:,:,-4:], box_size=box_size)
            loss = loss / accumulate_steps
            loss.backward()
            forward_pass_time = end_time - start_time
            running_loss += loss.item()
            forward_pass_times.append(forward_pass_time)

            # Do an optimizer step on accumulated gradients
            if (i+1) % accumulate_steps == 0:
                optimizer.step()
                optimizer.zero_grad()
                n_iter += 1

                # Log training loss every 200 optimizer steps
                if n_iter % 200 == 0:
                    writer.add_scalar('Loss/train', running_loss / 200, n_iter)
                    running_loss = 0.0

            # Decay learning rate every 200k steps with lower imit of 10^-7
            # (if smooth decay is not set)
            if ((n_iter % (2 * 10**5) == 0) and not smooth_lr_decay) and \
                                        (optimizer.param_groups[0]['lr'] > 10**(-7)):
                scheduler.step()
                # Log lr for tensorboard
                writer.add_scalar('LearningRate/optimizer_lr',
                                  optimizer.param_groups[0]['lr'], n_iter)

            # Save model after specified number of seconds (if set)
            if (save_after_time > 0):
                if ((time.monotonic() - start_time_for_save) > save_after_time):
                    torch.save(model.state_dict(), os.path.join(model_save_path, model_name))
                    start_time_for_save = time.monotonic()
        
        # Decay learning rate every epoch with lower imit of 10^-7 (if smooth decay is set)
        if (smooth_lr_decay) and (optimizer.param_groups[0]['lr'] > 10**(-7)):
            scheduler.step()
            # Log lr for tensorboard
            writer.add_scalar('LearningRate/optimizer_lr',
                              optimizer.param_groups[0]['lr'], n_iter)

        # Save model each epoch
        torch.save(model.state_dict(), os.path.join(model_save_path, model_name))

        # Save a checkpoint each epoch
        torch.save({'epoch': epoch+1, 'model_state_dict': model.state_dict(),
                    'optimizer_state_dict': optimizer.state_dict(),
                    'scheduler_state_dict': scheduler.state_dict()},
                   os.path.join(model_save_path, model_name + "_checkpoint.tar"))
            
        # Evaluate on validation set every validate_epochs epoch.
        # Always validate and save the model after the last epoch
        if (validate and (epoch % validate_epochs == 0)) or (epoch == epochs - 1) :
            model.eval()
            running_test_loss = 0.0
            with torch.no_grad():
                for j, test_data in enumerate(validation_loader, 0):
                    # Get validation data
                    inputs, targets = test_data
                    # Push to GPU
                    inputs = inputs.to(device, non_blocking=True)
                    targets = targets.to(device, non_blocking=True)
                    # Get outputs
                    outputs = model(inputs)
                    # Get loss
                    test_loss = PBC_MSE_loss(outputs, targets[:,:,-4:],
                                             box_size=box_size).item()
                    running_test_loss += test_loss
                    
            model.train()

            # Log validation loss for tensorboard
            validation_loss = running_test_loss / len(validation_loader)
            writer.add_scalar('Loss/validation', validation_loss, n_iter)


    writer.close()

    # Print validation loss at the end of the training
    print(f'Finished Training. Final validation loss is {validation_loss}')
    forward_pass_times = (np.array(forward_pass_times) / 1000000) # in ms instead of ns
    print(f'Forward step took {np.mean(forward_pass_times)} ms on average' +\
          f'(std: {np.std(forward_pass_times)})')
    # Return the filename of the model
    return model_name


def evaluate_model(model, model_file="", dataset="3_particles", model_dataset="",
                   model_dir="models", data_dir="Particle_Movement/data",
                   experiment_dir="", start_id=0, end_id=-1, target_step=1):

    # Set evaluation dataset as model dataset (dataset model was trained on)
    # if no model dataset was specified
    if len(model_dataset) < 1:
        model_dataset = dataset

    # Model path and output folder path
    if len(experiment_dir) > 0:
        model_path = os.path.join(model_dir, experiment_dir, model_dataset, model_file)
        output_folder_path = os.path.join(data_dir, experiment_dir, dataset,
                                          'test_predictions', model_file)
    else:
        model_path = os.path.join(model_dir, model_dataset, model_file)
        output_folder_path = os.path.join(data_dir, dataset, 'test_predictions',
                                          model_file)

    # Use batch_size=1 for inference
    batch_size = 1
    
    # Load test data set
    test_set = TrajectoryDataset(folder_path=os.path.join(data_dir, dataset),
                                 split='test', rollout=True, target_step=target_step)
    test_loader = torch.utils.data.DataLoader(test_set, batch_size=batch_size,
                                              shuffle=False)

    # Get parameters form dataset
    box_size = test_set.box_size
    time_step = test_set.time_step
    n_particles = test_set.n_particles
    physical_const = test_set.physical_const
    softening = test_set.softening
    softening_radius = test_set.softening_radius

    model.load_state_dict(torch.load(model_path))

    # Remove old output subfolder if it exists in case all trajectories will be built
    if start_id == 0 and end_id == -1:
        recreate_folder(output_folder_path)
    else:
        create_folder(output_folder_path)

    # Set proper end_id
    if end_id == -1:
        end_id = len(test_set)

    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    model.to(device)
    model.eval()

    # Build a tensor of step sizes for each sample in the batch
    dt = torch.Tensor([time_step]).to(device).unsqueeze(0).expand(batch_size, 1) * target_step

    # Log all trajectories for RMS over all trajectories
    predicted_trajectories = []

    print("Evaluationg model %s on %s test dataset" % (model_file, dataset))

    for i, data in enumerate(test_loader, 0):
        if start_id <= i < end_id:
            inputs, targets = data

            # Log the predicted trajecotry
            output_trajectory = np.zeros((inputs.shape[1]+1, inputs.shape[2], inputs.shape[3]))
            output_trajectory[0] = inputs[0][0].numpy()

            # Forward pass 
            current_state = inputs[0][0].unsqueeze(0)
            current_state = current_state.to(device)
            for j in range(inputs.shape[1]):
                output = model(current_state)
                # Detach to stop graph unroll in next loop iteration 
                current_state = torch.cat([current_state[:,:,:-4], output],
                                          dim = 2).detach() 
                # [timesteps, particles, state]; state = [m,x,y,v_x,v_y]
                output_trajectory[j+1, :, :] = current_state.cpu().detach().numpy()

            
            print("RMSE for trajectory %i: %f" % (i, pbc_rms_error(output_trajectory[1:,:,1:],
                                        targets.cpu().numpy()[0][:,:,1:], box_size=box_size)))
        
            # Save the predicted trajectory
            output_filename = os.path.join(output_folder_path,
                                           "predicted_trajectory_{i}.npy".format(i=i))
            np.save(output_filename, output_trajectory)

            # Log for RMS over all trajectories
            predicted_trajectories.append(output_trajectory)


    # RMS over all trajectories
    # [trajectories, timesteps, particles, state]; state = [m,x,y,v_x,v_y]
    predicted_trajectories = np.stack(predicted_trajectories, axis=0) 
    print("RMSE over all trajectories: %f" % 
          (pbc_rms_error(predicted_trajectories[:,1:,:,1:],
            test_set.trajectories[start_id:end_id,target_step::target_step,:,1:].numpy(),
            box_size=box_size)))
    print("Mean relative energy error over all trajectories: %f" % 
          (pbc_mean_relative_energy_error(predicted_trajectories, box_size=box_size,
                                    physical_const=physical_const,softening=softening,
                                    softening_radius=softening_radius)))

    print('Finished evaluation')

    # Return the output path
    return output_folder_path

def calculate_r2_from_single(old_state, new_state, predicted_new_state, box_size):
    
    true_delta = pbc_diff(new_state, old_state, box_size=box_size)
    predicted_delta = pbc_diff(predicted_new_state, old_state, box_size=box_size)
    true_delta_mean = torch.mean(true_delta, 0)
    ss_res = true_delta - predicted_delta
    ss_res = ss_res**2
    ss_res = ss_res.sum().item()
    ss_tot = true_delta - true_delta_mean
    ss_tot = ss_tot**2
    ss_tot = ss_tot.sum().item()
    r2_coefficient = 1 - ss_res/ss_tot
    
    return r2_coefficient

def calculate_avg_r2(model, model_file="", dataset="3_particles", model_dataset="",
                     model_dir="models", data_dir='Particle_Movement\data', experiment_dir="",
                     start_id=0, end_id=-1, target_step=1):
    
    # Set evaluation dataset as model dataset (dataset model was trained on)
    # if no model dataset was specified
    if len(model_dataset) < 1:
        model_dataset = dataset

    # Model path and output folder path
    if len(experiment_dir) > 0:
        model_path = os.path.join(model_dir, experiment_dir, model_dataset, model_file)
        output_folder_path = os.path.join(data_dir, experiment_dir, dataset,
                                          'test_predictions', model_file)
    else:
        model_path = os.path.join(model_dir, model_dataset, model_file)
        output_folder_path = os.path.join(data_dir, dataset, 'test_predictions',
                                          model_file)

    # Use batch_size=1 for inference
    batch_size = 1
    
    # Load test data set
    test_set = TrajectoryDataset(folder_path=os.path.join(data_dir, dataset),
                                 split='test', rollout=True, target_step=target_step)
    test_loader = torch.utils.data.DataLoader(test_set, batch_size=batch_size,
                                              shuffle=False)

    # Get parameters form dataset
    box_size = test_set.box_size

    model.load_state_dict(torch.load(model_path))

    # Remove old output subfolder if it exists in case all trajectories will be built
    if start_id == 0 and end_id == -1:
        recreate_folder(output_folder_path)
    else:
        create_folder(output_folder_path)

    # Set proper end_id
    if end_id == -1:
        end_id = len(test_set)

    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    model.to(device)
    model.eval()

    print("Evaluationg model %s on %s test dataset" % (model_file, dataset))

    total_r2 = []
    for i, data in enumerate(test_loader, 0):
        if start_id <= i < end_id:
            inputs, targets = data

            # Log the predicted trajecotry
            output_trajectory = np.zeros((inputs.shape[1]+1, inputs.shape[2], inputs.shape[3]))
            output_trajectory[0] = inputs[0][0].numpy()

            # Forward pass 
            r2_coeffs = []
            for j in range(inputs.shape[1] - 1):
                current_state = inputs[0][j].unsqueeze(0)
                current_state = current_state.to(device)
                output = model(current_state)
                
                single_r2_coeff = calculate_r2_from_single(inputs[0][j][:,-4:].to(device),
                                        inputs[0][j+1][:,-4:].to(device), output, box_size)
                r2_coeffs.append(single_r2_coeff)
                
            trajectory_r2 = np.mean(r2_coeffs)
            total_r2.append(trajectory_r2)
            
            print("R2 for trajectory %i: %f" % (i, trajectory_r2))

    print("R2 over all trajectories: %f" % (np.mean(total_r2)))

    print('Finished R2 calculation')

    # Return the output path
    return output_folder_path