#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Oct 25 20:10:37 2021

@author: sandro
"""

import torch
from torch.distributions.categorical import Categorical
from torch.nn.functional import normalize
import sys

import numpy as np
import os
from time import time

from omegaconf import OmegaConf
from torch_points3d.datasets.segmentation import ShapeNetDataset

device = "cuda" if torch.cuda.is_available() else "cpu"
print("Using {} device".format(device))

data_augmentation_yaml = """
RandomGaussianNoise:
    params:
        sigma: 0.01
        clip: 0.05
RandomRotation:
    params:
        max_degrees: 180
RandomAnisotropicScaling:
    params:
        min: 0.9
        max: 1.1
"""

pi = torch.tensor([3.1415626]).to(device)

data_augmentation = OmegaConf.create(data_augmentation_yaml)

class ShapeNet_Dataloader():
    
    seg_classes = {
        "Airplane": [0, 1, 2, 3],
        "Bag": [4, 5],
        "Cap": [6, 7],
        "Car": [8, 9, 10, 11],
        "Chair": [12, 13, 14, 15],
        "Earphone": [16, 17, 18],
        "Guitar": [19, 20, 21],
        "Knife": [22, 23],
        "Lamp": [24, 25, 26, 27],
        "Laptop": [28, 29],
        "Motorbike": [30, 31, 32, 33, 34, 35],
        "Mug": [36, 37],
        "Pistol": [38, 39, 40],
        "Rocket": [41, 42, 43],
        "Skateboard": [44, 45, 46],
        "Table": [47, 48, 49],
    }
    
    remainder_indices = torch.tensor([[0,0,1,1,2,2],[1,2,0,2,0,1]]).to(device)
    normal_indices = torch.tensor([2,1,2,0,1,0]).to(device)
    remainder_signs = torch.tensor([[1,-1,1],
                                    [1,1,-1],
                                    [-1,1,1]]).to(device)
    
    def __init__(self, directory, category='All', use_normals=False,
                 data_augmentation=data_augmentation):
        
        shapenet_yaml = """
        class: shapenet.ShapeNetDataset
        task: segmentation
        dataroot: %s
        normal: %r                                    # Use normal vectors as features
        first_subsampling: 0.02                       # Grid size of the input data
        pre_transforms:                               # Offline transforms, done only once
            - transform: NormalizeScale           
            - transform: GridSampling3D
              params:
                size: ${first_subsampling}
        """ % (os.path.join(directory,"data"), use_normals) 
        
        params = OmegaConf.create(shapenet_yaml)
        if category != "All":
            params.category = category
            self.min_label = self.seg_classes[category][0]
            self.max_label = self.seg_classes[category][-1]
        else:
            self.min_label = 0
            self.max_label = 49
        self.label_count = self.max_label - self.min_label + 1
        self.directory = directory
        self.use_normals = use_normals
        self.dataset = ShapeNetDataset(params)
        
        self.parse_data_augmentation(data_augmentation)
        
        if category == "All":
            train_sample_probabilities =\
                self.compute_sample_probabilities(self.dataset.train_dataset)
        else:
            length = len(self.dataset.train_dataset)
            prob = 1.0/length
            train_sample_probabilities = torch.tensor(length*[prob]).to(device)
            
        self.train_sample_distribution = Categorical(train_sample_probabilities)
        self.train_dataset_size = train_sample_probabilities.shape[0]
        
        self.test_sample_factors = self.compute_test_dataset_factors(
                                                self.dataset.test_dataset)
        self.test_points = self.compute_test_dataset_points(
                                                self.dataset.test_dataset)
        self.test_labels = self.compute_test_dataset_labels(
                                                self.dataset.test_dataset)
        
        self.val_sample_factors = self.compute_val_dataset_factors(
                                                self.dataset.val_dataset)
        self.val_points = self.compute_val_dataset_points(
                                                self.dataset.val_dataset)
        self.val_labels = self.compute_val_dataset_labels(
                                                self.dataset.val_dataset)
        
    def parse_data_augmentation(self, data_augmentation):
        
        if 'RandomGaussianNoise' in data_augmentation:
            self.gaussian_noise = True
            try:
                self.gaussian_noise_sigma =\
                    data_augmentation['RandomGaussianNoise']['params']['sigma']
                self.gaussian_noise_clip =\
                    data_augmentation['RandomGaussianNoise']['params']['clip']
            except ValueError:
                print("Could not convert given random gaussian noise parameters. " +\
                      "Define \'sigma\' and \'clip\'!")
        else:
            self.gaussian_noise = False
            
        if 'RandomRotation' in data_augmentation:
            self.rotation = True
            try:
                self.rotation_max_degrees =\
                    data_augmentation['RandomRotation']['params']['max_degrees']/180.0*pi
            except ValueError:
                print("Could not convert given random rotation parameter. " +\
                      "Define \'max_degrees\'!")
        else:
            self.rotation = False
            
        if 'RandomAnisotropicScaling' in data_augmentation:
            self.anisotropic_scaling = True
            try:
                self.anisotropic_scaling_min =\
                    data_augmentation['RandomAnisotropicScaling']['params']['min']
                self.anisotropic_scaling_max =\
                    data_augmentation['RandomAnisotropicScaling']['params']['max']
                self.anisotropic_scaling_factor = self.anisotropic_scaling_max -\
                                                    self.anisotropic_scaling_min
            except ValueError:
                print("Could not convert given random anisotropic scaling parameters. " +\
                      "Define \'min\' and \'max\'!")
        else:
            self.anisotropic_scaling = False
            
    def compute_sample_probabilities(self, train_dataset):
        
        num_categories = len(train_dataset.categories)
        category_probabilities = torch.tensor(num_categories*
                                              [1.0/num_categories]).to(device)
        
        category_counts = torch.tensor(num_categories*[0]).to(device)
        for sample in train_dataset:
            category_counts[sample.category[0].item()] += 1
        
        for category in range(len(category_probabilities)):
            category_probabilities[category] /= category_counts[category]
        
        sample_probabilities = torch.repeat_interleave(category_probabilities,
                                                       category_counts)
        
        return sample_probabilities
    
    def compute_test_dataset_factors(self, test_dataset):
        
        num_categories = len(test_dataset[0].categories)
        
        category_counts = torch.tensor(num_categories*[0]).to(device)
        for sample in test_dataset[0]:
            category_counts[sample.category[0].item()] += 1
        
        test_dataset_category_factors = torch.div(1,
                                                  category_counts)
        test_dataset_category_factors = torch.div(test_dataset_category_factors,
                                                  len(test_dataset[0].categories))
        
        test_dataset_factors = torch.zeros(len(test_dataset[0])).to(device)
        i = 0
        for sample in test_dataset[0]:
            test_dataset_factors[i] = test_dataset_category_factors[
                                                    sample.category[0].item()]
            i += 1
        
        return test_dataset_factors
    
    def compute_val_dataset_factors(self, val_dataset):
        
        num_categories = len(val_dataset.categories)
        
        category_counts = torch.tensor(num_categories*[0]).to(device)
        for sample in val_dataset:
            category_counts[sample.category[0].item()] += 1
        
        val_dataset_category_factors = torch.div(1,
                                                 category_counts)
        val_dataset_category_factors = torch.div(val_dataset_category_factors,
                                                 len(val_dataset.categories))
        
        val_dataset_factors = torch.zeros(len(val_dataset)).to(device)
        i = 0
        for sample in val_dataset:
            val_dataset_factors[i] = val_dataset_category_factors[
                                                    sample.category[0].item()]
            i += 1
        
        return val_dataset_factors
    
    def compute_test_dataset_points(self, test_dataset):
        
        test_dataset_points = []
        for sample in test_dataset[0]:
            if self.use_normals:
                pos = sample.pos
                normals = sample.x
                points = torch.cat((pos, normals), dim=-1)
            else:
                points = sample.pos
            test_dataset_points.append(points.to(device))
        
        return test_dataset_points
    
    def compute_val_dataset_points(self, val_dataset):
        
        val_dataset_points = []
        for sample in val_dataset:
            if self.use_normals:
                pos = sample.pos
                normals = sample.x
                points = torch.cat((pos, normals), dim=-1)
            else:
                points = sample.pos
            val_dataset_points.append(points.to(device))
        
        return val_dataset_points
    
    def compute_test_dataset_labels(self, test_dataset):
        
        test_dataset_labels = []
        for sample in test_dataset[0]:
            labels = sample.y
            test_dataset_labels.append(labels.to(device))
        
        return test_dataset_labels
    
    def compute_val_dataset_labels(self, val_dataset):
        
        val_dataset_labels = []
        for sample in val_dataset:
            labels = sample.y
            val_dataset_labels.append(labels.to(device))
        
        return val_dataset_labels
            
    def tensorize_set(self, batch, batch_size):
        
        N_min = batch[0].y.shape[0]
        N_max = batch[0].y.shape[0]
        for sample in batch:
            if sample.y.shape[0] < N_min:
                N_min = sample.y.shape[0]
            if sample.y.shape[0] > N_max:
                N_max = sample.y.shape[0]
                
        N = torch.randint(N_min, N_max+1, [1], device=device)
        if N > 2*N_min:
            N = 2*N_min
        
        if self.use_normals:
            batch_points = torch.zeros((batch_size, N, 6), device=device)
        else:
            batch_points = torch.zeros((batch_size, N, 3), device=device)
        batch_labels = torch.zeros((batch_size, N), dtype=torch.long, device=device)
        
        sample_idx = 0
        for sample in batch:
            N_sample = sample.y.shape[0]
            N_difference = N - N_sample
            if N_difference > 0:
                indices = torch.randperm(N_sample, device=device)
                indices = torch.cat((indices, indices[:N_difference]))
            elif N_difference < 0:
                indices = torch.randperm(N_sample, device=device)
                indices = indices[:N_difference]
            else:
                indices = torch.randperm(N_sample, device=device)
                
            points = sample.pos[indices]
            if self.use_normals:
                normals = sample.x
                normals = normals[indices]
                batch_points[sample_idx] = torch.cat((points, normals), 1)
            else:
                batch_points[sample_idx] = points
            batch_labels[sample_idx] = sample.y[indices]
                    
            sample_idx += 1
        
        return batch_points, batch_labels
    
    def random_anisotropic_scaling(self, batch_points):
        
        batch_size = batch_points.shape[0]
        
        direction = torch.randn((batch_size, 3), device=device)
        direction = normalize(direction, dim=-1)
        angle = torch.mul(torch.rand([batch_size], device=device), pi)
        scaling_factors = torch.tensor(batch_size*[3*[self.anisotropic_scaling_min]]\
            ).to(device) + torch.mul(torch.rand([batch_size, 3], device=device),
                                     self.anisotropic_scaling_factor)
        
        # rotate
        batch_points = self.rotate(batch_points, direction, angle)
        # scale anisotropically
        if self.use_normals:
            scaling_factors = torch.cat((scaling_factors, scaling_factors), dim=1)
        batch_points = torch.mul(scaling_factors.unsqueeze(1), batch_points)
        # rotate back
        batch_points = self.rotate(batch_points, direction, -angle)
        
        if self.use_normals:
            batch_points = self.rescale_normals(batch_points)
        
        return batch_points
    
    def random_rotation(self, batch_points):
        
        batch_size = batch_points.shape[0]
        
        direction = torch.randn((batch_size, 3), device=device)
        direction = normalize(direction, dim=-1)
        angle = torch.mul(torch.rand([batch_size], device=device),
                          self.rotation_max_degrees)
        
        batch_points = self.rotate(batch_points, direction, angle)
        
        return batch_points
        
    def random_gaussian_noise(self, batch_points):
        
        noise = torch.randn(batch_points.shape, device=device).to(device)
        noise = torch.mul(self.gaussian_noise_sigma, noise)
        noise = torch.clip(noise, min=-self.gaussian_noise_clip,
                           max=self.gaussian_noise_clip)
        batch_points = torch.add(batch_points, noise)
        
        if self.use_normals:
            batch_points = self.rescale_normals(batch_points)
        
        return batch_points
    
    def rescale_normals(self, batch_points):
        
        points = batch_points[:,:,:3]
        normals = batch_points[:,:,3:]
        
        normals = normalize(normals, dim=-1)
        
        batch_points = torch.cat((points, normals), dim=-1)
        
        return batch_points
    
    def rotate(self, batch_points, direction, angle):
        
        batch_size = batch_points.shape[0]
        
        sina = torch.sin(angle)
        cosa = torch.cos(angle)
        
        ddt = torch.matmul(direction.unsqueeze(2), direction.unsqueeze(1))
        ddt = torch.mul((1 - cosa).unsqueeze(-1).unsqueeze(-1), ddt)
        
        remainder = torch.zeros((batch_size, 3, 3), device=device)
        remainder[:,self.remainder_indices[0],self.remainder_indices[1]] =\
                                                direction[:,self.normal_indices]
        remainder = torch.mul(sina.unsqueeze(-1).unsqueeze(-1), remainder)
        remainder = torch.mul(self.remainder_signs, remainder)
        
        cosa_matrix = torch.mul(cosa.unsqueeze(-1).unsqueeze(-1),
                                torch.eye(3, device=device).unsqueeze(0))
        
        rotation_matrix = torch.add(ddt, remainder)
        rotation_matrix = torch.add(rotation_matrix, cosa_matrix)
        
        if self.use_normals:
            points = batch_points[:,:,:3]
            normals = batch_points[:,:,3:]
            points = torch.matmul(rotation_matrix,
                                  points.transpose(-1, -2)).transpose(-1, -2)
            normals = torch.matmul(rotation_matrix,
                                   normals.transpose(-1, -2)).transpose(-1, -2)
            batch_points = torch.cat((points, normals), dim=-1)
        else:
            batch_points = torch.matmul(rotation_matrix,
                                        batch_points.transpose(-1, -2)).transpose(-1, -2)
        
        return batch_points
    
    def train_pipeline(self, train_batch_points):
        
        if self.anisotropic_scaling:
            train_batch_points = self.random_anisotropic_scaling(train_batch_points)
            train_batch_points = self.rescale_normals(train_batch_points)
        if self.rotation:
            train_batch_points = self.random_rotation(train_batch_points)
        if self.gaussian_noise:
            train_batch_points = self.random_gaussian_noise(train_batch_points)
            
        return train_batch_points

    def get_train_batch(self, batch_size):
        
        samples = self.train_sample_distribution.sample([batch_size])
        batch = []
        for sample in samples:
            batch.append(self.dataset.train_dataset[sample])
        
        train_batch_points, train_batch_labels = self.tensorize_set(batch, batch_size)
        train_batch_labels -= self.min_label
        
        train_batch_points = self.train_pipeline(train_batch_points)
        
        return train_batch_points, train_batch_labels
    
    def get_test_dataset(self):
        
        return self.test_points, np.subtract(self.test_labels, self.min_label
                                             ).tolist(), self.test_sample_factors
    
    def get_val_dataset(self):
        
        return self.val_points, np.subtract(self.val_labels, self.min_label
                                            ).tolist(), self.val_sample_factors
    
    def get_label_count(self):
        
        return self.label_count
    
    def get_train_dataset_size(self):
        
        return self.train_dataset_size
    
    def get_point_cloud_statistics(self):
        
        train_dataset_size = self.train_dataset_size
        validation_dataset_size = len(self.dataset.val_dataset)
        test_dataset_size = len(self.dataset.test_dataset[0])
        # test_dataset_size = len(self.test_points)
        
        total_size = train_dataset_size + validation_dataset_size + test_dataset_size
        
        min_points = self.dataset.train_dataset[0].category.shape[0]
        max_points = self.dataset.train_dataset[0].category.shape[0]
        print('train dataset 0')
        print(self.dataset.train_dataset[1].category.shape[0])
        avg_points = 0
        for dataset in [self.dataset.train_dataset, self.dataset.test_dataset[0],
                        self.dataset.val_dataset]:
            for point_cloud in dataset:
                min_points = min(point_cloud.category.shape[0], min_points)
                max_points = max(point_cloud.category.shape[0], max_points)
                avg_points += point_cloud.category.shape[0]/total_size
                
        return min_points, max_points, avg_points, train_dataset_size,\
            validation_dataset_size, test_dataset_size

if __name__ == '__main__':
    
    batch_points = torch.eye(3)
    batch_points = torch.stack((batch_points, batch_points, batch_points), dim=0)
    print(batch_points)
    batch_points = torch.cat((batch_points, batch_points), dim=-1)
    print(batch_points)
    
    direction = torch.tensor([[1,1,0],[1,0,1],[0,1,1]], dtype=torch.float).to(device)
    direction = normalize(direction, dim=-1)
    angle = torch.tensor([0.5*pi, 0.5*pi, 0.5*pi]).to(device)
    angle = torch.tensor([pi, pi, pi]).to(device)
    
    directory = 'tst_datasets'
    
    dataloader = ShapeNet_Dataloader(directory, category='Car', use_normals=True)
    min_points, max_points, avg_points, train_dataset_size,\
        validation_dataset_size, test_dataset_size = dataloader.get_point_cloud_statistics()
    print('Dataset Statistics:')
    print('min points: ' + str(min_points))
    print('max points: ' + str(max_points))
    print('avg points: ' + str(avg_points))
    print('train dataset size: ' + str(train_dataset_size))
    print('validation dataset size: ' + str(validation_dataset_size))
    print('test dataset size: ' + str(test_dataset_size))
    
    points, labels = dataloader.get_train_batch(10)
    print(points)
    print(labels)
    # batch_points = dataloader.rotate(batch_points, direction, angle)
    # batch_points = dataloader.random_anisotropic_scaling(batch_points)
    # batch_points = dataloader.random_rotation(batch_points)
    # batch_points = dataloader.random_gaussian_noise(batch_points)
    batch_points = dataloader.train_pipeline(batch_points)
    
    print(batch_points)
    
    points, labels, factors = dataloader.get_test_dataset()
    print(points[0])
    print(labels[0])
    print(factors)