#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Feb  3 17:41:04 2022

@author: sandro
"""

import argparse
import logging
import os
os.environ['WANDB_MODE'] = 'offline'
import math
import copy
import torch
from dataclasses import dataclass, field
from transformers import RobertaForMaskedLM, RobertaTokenizerFast, TextDataset,\
    DataCollatorForLanguageModeling, Trainer
from transformers import TrainingArguments, HfArgumentParser
#from transformers.modeling_longformer import LongformerSelfAttention

from TreeBert.Tree_Attention import TreeRobertaForMaskedLM, create_long_model

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)

# parse command line arguments
parser = argparse.ArgumentParser(description='Parameters for RoBERTA '+\
                                 'trials')
parser.add_argument('-model_dir', metavar='model_dir', type=str,
                    default='temp_models', help='Where to save the model')
parser.add_argument('-model_name', metavar='model_name', type=str,
                    default='DagoBERT', help='Model name')
parser.add_argument('-max_seq_len', metavar='max_seq_len', type=int, default=512,
                    help='Maximum sequence length the model should be trained on')
parser.add_argument('-tree_depth', metavar='tree_depth', type=int, default=5,
                    help='The depth of the attention tree')
parser.add_argument('-max_steps', metavar='max_steps', type=int, default=5000,
                    help='Maximum pretraining gradient updates')
parser.add_argument('-model', metavar='model', type=str, default='treeBERTa',
                    help='use treeBERTa for tree attention and RoBERTa for dense att.')

args = parser.parse_args()
model_dir = args.model_dir
model_name = args.model_name
max_seq_len = args.max_seq_len
tree_depth = args.tree_depth
max_steps = args.max_steps
model_architecture = args.model

def pretrain_and_evaluate(args, model, tokenizer, eval_only, model_path):
    val_dataset = TextDataset(tokenizer=tokenizer,
                              file_path=args.val_datapath,
                              block_size=tokenizer.max_len)
    if eval_only:
        train_dataset = val_dataset
    else:
        logger.info(
            f'Loading and tokenizing training data is usually slow: {args.train_datapath}')
        train_dataset = TextDataset(tokenizer=tokenizer,
                                    file_path=args.train_datapath,
                                    block_size=tokenizer.max_len)

    data_collator = DataCollatorForLanguageModeling(tokenizer=tokenizer,
                                                    mlm=True, mlm_probability=0.15)
    trainer = Trainer(model=model, args=args, data_collator=data_collator,
                      train_dataset=train_dataset, eval_dataset=val_dataset,
                      prediction_loss_only=True,)

    eval_loss = trainer.evaluate()
    eval_loss = eval_loss['eval_loss']
    logger.info(f'Initial eval bpc: {eval_loss/math.log(2)}')
    
    if not eval_only:
        print('starting training...')
        trainer.train(model_path=model_path)
        trainer.save_model()

        eval_loss = trainer.evaluate()
        eval_loss = eval_loss['eval_loss']
        logger.info(f'Eval bpc after pretraining: {eval_loss/math.log(2)}')
        
@dataclass
class ModelArgs:
    max_pos: int = field(default=512, metadata={"help": "Maximum position"})

parser = HfArgumentParser((TrainingArguments, ModelArgs,))


training_args, model_args = parser.parse_args_into_dataclasses(
    look_for_args_file=False, args=[
    '--output_dir', model_dir,
    '--warmup_steps', '500',
    '--learning_rate', '0.00003',
    '--weight_decay', '0.01',
    '--adam_epsilon', '1e-6',
    '--max_steps', str(max_steps),
    '--logging_steps', '50',
    '--save_steps', '500',
    '--max_grad_norm', '5.0',
    '--per_gpu_eval_batch_size', '1',
    '--per_gpu_train_batch_size', '8',  # 32GB gpu with fp32
    '--gradient_accumulation_steps', '8',
    '--evaluate_during_training',
    '--do_train',
    '--do_eval',
])
training_args.val_datapath = 'TreeBert/wikitext-103-raw/wiki.valid.raw'
training_args.train_datapath = 'TreeBert/wikitext-103-raw/wiki.train_short.raw'
            
if __name__ == '__main__':
    
    model_args.max_pos = max_seq_len
    
    model_path = f'{training_args.output_dir}/{model_name}-{model_args.max_pos}'
    if not os.path.exists(model_path):
        os.makedirs(model_path)
    
    if model_architecture == 'treeBERTa':
        logger.info(f'Converting roberta-base into {model_name}-{model_args.max_pos}')
        model, tokenizer = create_long_model(
            save_model_to=model_path, max_pos=model_args.max_pos, tree_depth=tree_depth,
            logger=logger)
    elif model_architecture == 'RoBERTa':
        logger.info('using roberta-base')
        model = RobertaForMaskedLM.from_pretrained('roberta-base')
        tokenizer = RobertaTokenizerFast.from_pretrained('roberta-base',
                                                         model_max_length=512)
        logger.info(f'saving model to {model_path}')
        model.save_pretrained(model_path)
        tokenizer.save_pretrained(model_path)
    
    logger.info(f'Loading the model from {model_path}')
    tokenizer = RobertaTokenizerFast.from_pretrained(model_path)
    if model_architecture == 'treeBERTa':
        model = TreeRobertaForMaskedLM.from_pretrained(model_path)
    elif model_architecture == 'RoBERTa':
        model = RobertaForMaskedLM.from_pretrained(model_path)
    
    logger.info(f'Pretraining tree-roberta-base-{model_args.max_pos} ... ')
    
    pretrain_and_evaluate(training_args, model, tokenizer, eval_only=False,
                          model_path=training_args.output_dir)
    
    