#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Jan 31 18:19:44 2022

@author: sandro
"""

import math
import torch
import torch.nn as nn
import sys

from transformers import RobertaForMaskedLM, RobertaTokenizerFast
# from transformers.modeling_roberta import RobertaConfig, RobertaModel
from transformers import RobertaModel

sys.path.append('..')

from Tree_attention_scaling_simple_tree_no_sparse.tree_attention_modules_simple_tree\
    import Simple_Attention_Tree
    
device = "cuda" if torch.cuda.is_available() else "cpu"
print("Using {} device".format(device))

class Tree_Attention(nn.Module):
    
    def __init__(self,
                 config,
                 scaled_softmax = True,
                 max_split = True,
                 tree='plusplus',
                 iterations = 5,
                 noise_coeff = 0.5):
        
        super().__init__()
        
        self.depth = config.tree_depth
        self.n_leafs = int(math.pow(2, self.depth))
        
        self.embed_dim = config.hidden_size
        self.num_heads = config.num_attention_heads
        self.head_dim = int(config.hidden_size / config.num_attention_heads)
        
        self.tree_attention = Simple_Attention_Tree(self.n_leafs, self.head_dim,
                                scaled_softmax=scaled_softmax, max_split=max_split,
                                tree=tree, tree_plusplus_iterations=iterations,
                                tree_plusplus_noise_coefficient=noise_coeff)
        
        self.query = nn.Linear(config.hidden_size, self.embed_dim)
        self.key = nn.Linear(config.hidden_size, self.embed_dim)
        self.value = nn.Linear(config.hidden_size, self.embed_dim)

        self.dropout = nn.Dropout(config.attention_probs_dropout_prob)
        
    def forward(self,
                hidden_states,
                attention_mask=None,
                head_mask=None,
                encoder_hidden_states=None,
                encoder_attention_mask=None,
                past_key_value=None,
                output_attentions=False):
        
        assert encoder_hidden_states is None,\
            "`encoder_hidden_states` is not supported and should be None"
        assert encoder_attention_mask is None,\
            "`encoder_attention_mask` is not supported and should be None"
        assert output_attentions is False,\
            "`output_attentions` is not supported and should be None"
        assert attention_mask.sum() == 0,\
            "`attention_mask` is not supported and should be None"
        
        mixed_query_layer = self.query(hidden_states)
        mixed_key_layer = self.key(hidden_states)
        mixed_value_layer = self.value(hidden_states)

        mixed_query_layer = torch.cat(mixed_query_layer.split(self.head_dim, 2), 0)
        mixed_key_layer = torch.cat(mixed_key_layer.split(self.head_dim, 2), 0)
        mixed_value_layer = torch.cat(mixed_value_layer.split(self.head_dim, 2), 0)

        output = self.tree_attention(mixed_key_layer, mixed_query_layer,
                                     mixed_value_layer)
        output = torch.cat(output.split(hidden_states.size(0), 0), 2)
        
        return (output, )

class TreeRobertaForMaskedLM(RobertaForMaskedLM):
    def __init__(self, config):
        super().__init__(config)
        for i, layer in enumerate(self.roberta.encoder.layer):
            # replace the `modeling_bert.BertSelfAttention` object with
            # Tree Attention
            layer.attention.self = Tree_Attention(config)
            
class TreeRoberta(RobertaModel):
    def __init__(self, config):
        super(TreeRoberta, self).__init__(config)
        for i, layer in enumerate(self.encoder.layer):
            # replace the `modeling_bert.BertSelfAttention` object with
            # Tree Attention
            layer.attention.self = Tree_Attention(config)
        
def create_long_model(save_model_to, max_pos, tree_depth, logger):
    model = RobertaForMaskedLM.from_pretrained('roberta-base')
    tokenizer = RobertaTokenizerFast.from_pretrained('roberta-base',
                                                     model_max_length=max_pos)
    config = model.config

    # extend position embeddings
    current_max_pos, embed_size = model.roberta.embeddings.position_embeddings.weight.shape
    if max_pos > current_max_pos:
        tokenizer.model_max_length = max_pos
        tokenizer.init_kwargs['model_max_length'] = max_pos
        max_pos += 2  # NOTE: RoBERTa has positions 0,1 reserved, so embedding
                        # size is max position + 2
        config.max_position_embeddings = max_pos
        assert max_pos > current_max_pos
        # allocate a larger position embedding matrix
        new_pos_embed = model.roberta.embeddings.position_embeddings.weight.new_empty(
            max_pos, embed_size)
        # copy position embeddings over and over to initialize the new position embeddings
        k = 2
        step = current_max_pos - 2
        while k < max_pos - 1:
            new_pos_embed[k:(k + step)] =\
                model.roberta.embeddings.position_embeddings.weight[2:]
            k += step
        model.roberta.embeddings.position_embeddings.weight.data = new_pos_embed
    
    config.tree_depth = tree_depth
    # replace the `modeling_bert.BertSelfAttention` object with `LongformerSelfAttention`
    for i, layer in enumerate(model.roberta.encoder.layer):
        tree_attention = Tree_Attention(config)
        tree_attention.query = layer.attention.self.query
        tree_attention.key = layer.attention.self.key
        tree_attention.value = layer.attention.self.value

        layer.attention.self = tree_attention

    logger.info(f'saving model to {save_model_to}')
    model.save_pretrained(save_model_to)
    tokenizer.save_pretrained(save_model_to)
    return model, tokenizer
