#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Oct 29 14:33:06 2021

@author: sandro
"""
import torch
import torch.nn as nn
import torch.nn.functional as F
import math

from time import time, sleep
from torch.autograd import Function
from torch.distributions.categorical import Categorical
from torch.distributions.normal import Normal

device = "cuda" if torch.cuda.is_available() else "cpu"
print("Using {} device".format(device))

do_timing = False

class Tree_PlusPlus_Builder(nn.Module):
    
    def __init__(self, depth:int, dimension:int, iterations:int = 5,
                 noise_coeff:float = 0.5):
        super(Tree_PlusPlus_Builder, self).__init__()
        
        self.depth = depth
        self.n_leafs = int(math.pow(2, depth))
        self.device = device
        self.dimension = dimension
        
        self.iterations = iterations
        self.noise_coeff = torch.tensor([noise_coeff], device=self.device)
        self.noise_distribution = Normal(torch.tensor([0.0], device=self.device),
                                         self.noise_coeff)
    
    # initialize same as k-means++
    def initialize_projection_vectors(self, K):
        
        batch_size = K.shape[0]
        set_size = K.shape[1]
        batch_tensor = torch.arange(0, batch_size, 1, device=self.device, dtype=torch.long)
        
        # randomly select one point in each set
        first_chosen_inds = torch.randint(0, set_size, [batch_size,], dtype=torch.long)
        first_chosen_points = K[batch_tensor, first_chosen_inds]
        distances = torch.sub(K, first_chosen_points.unsqueeze(-2))
        distances = torch.add(torch.linalg.norm(distances, dim=-1), 1e-8)
        normalization_constants = torch.sum(distances, dim=-1).unsqueeze(-1)
        probabilities = torch.div(distances, normalization_constants)
        del distances, normalization_constants, first_chosen_inds
        
        # probabilistically select a second point in each set with probability
        # proportional to the distance to the first point
        second_point_distribution = Categorical(probs=probabilities)
        second_chosen_inds = second_point_distribution.sample()
        second_chosen_points = K[batch_tensor, second_chosen_inds]
        del second_point_distribution, second_chosen_inds
        
        projection_vectors = torch.sub(first_chosen_points, second_chosen_points)
        projection_vectors_norm = torch.linalg.norm(projection_vectors, dim=-1).unsqueeze(-1)
        projection_vectors = torch.div(projection_vectors, projection_vectors_norm)
        del projection_vectors_norm
        
        return projection_vectors.unsqueeze(-2)
    
    def split(self, K, K_indices, level):
        
        batch_size = K.shape[0]
        set_size = K.shape[1]
        noise_size_tensor = torch.tensor([batch_size, set_size], dtype=torch.int32,
                                         device=device)
        
        new_set_size = torch.floor(torch.div(set_size, 2)).int()
        new_batch_size = torch.mul(batch_size, 2).int()
        batch_inds = torch.arange(batch_size, device=self.device).repeat_interleave(
                                        set_size).reshape([batch_size, set_size])
        
        projection_vectors = self.initialize_projection_vectors(K)
        
        # through several iterations find a close to optimal projection direction
        for iteration in range(self.iterations):
            projection_results = torch.mul(K, projection_vectors).sum(dim=-1)
            rand_noise = self.noise_distribution.sample(noise_size_tensor).squeeze(-1)
            projection_results = projection_results + rand_noise
            del rand_noise
            sorted_inner_products, sort_indices = torch.sort(projection_results, dim=-1)
            del projection_results, sorted_inner_products, projection_vectors
            
            first_indices = sort_indices[:, :new_set_size]
            second_indices = sort_indices[:, new_set_size:]
            first_set = K[batch_inds[:, :new_set_size], first_indices]
            second_set = K[batch_inds[:, new_set_size:], second_indices]
            first_centers = torch.mean(first_set, dim=-2)
            second_centers = torch.mean(second_set, dim=-2)
            del first_indices, second_indices, first_set, second_set
            
            projection_vectors = torch.sub(second_centers, first_centers)
            projection_vectors_norm = torch.linalg.norm(projection_vectors, dim=-1).unsqueeze(-1)
            projection_vectors = torch.div(projection_vectors, projection_vectors_norm)
            projection_vectors = projection_vectors.unsqueeze(-2)
            del first_centers, second_centers
            
        # project each point into 1D and sort
        projection_results = torch.mul(K, projection_vectors).sum(dim=-1)
        sorted_inner_products, sort_indices = torch.sort(projection_results, dim=-1)
        del projection_results, sorted_inner_products, projection_vectors
        
        K = K[batch_inds, sort_indices]
        K = K.reshape([new_batch_size, new_set_size, K.shape[2]])
        K_indices = K_indices[batch_inds, sort_indices]
        K_indices = K_indices.reshape([new_batch_size, new_set_size])
        
        return K, K_indices
        
    def forward(self, K, batch_size:int, set_size:int):
        
        # in order to perform meaningful probabilistic splitting, the norm of
        # all Keys needs to be normalized
        K_norms = torch.linalg.norm(K, dim=-1)
        K_norms_mean = torch.mean(K_norms, dim=-1).unsqueeze(-1).unsqueeze(-1)
        K = torch.div(K, torch.add(K_norms_mean,
                                   torch.tensor([1e-12], device=self.device)))
    
        # pad the Key vectors such that the modified K can be evenly
        # distributed among the tree leafs
        n_padding_vectors = torch.remainder(torch.tensor([-set_size], device=self.device),
                                            self.n_leafs)
        Ks_per_bucket = math.ceil(set_size/self.n_leafs)
        padded_set_size = Ks_per_bucket*self.n_leafs
        # duplicate a random selection of key vectors to obtain padding vectors
        padding_vector_indices = torch.randint(0, set_size, [n_padding_vectors],
                                               dtype=torch.long, device=self.device)
        padding_vectors = K[:, padding_vector_indices]
        K = torch.cat((K, padding_vectors), dim=1)
        del padding_vectors, n_padding_vectors, padding_vector_indices
        K_indices = torch.arange(padded_set_size, device=self.device).repeat(
                                batch_size).reshape((batch_size, padded_set_size))
        
        # assign key vectors to tree leafs
        for level in torch.arange(self.depth, device=self.device):
            K, K_indices = self.split(K, K_indices, level)
        del K
            
        # find the leaf indices according to the reordered K_indices
        buckets = torch.arange(self.n_leafs, dtype=torch.int32, device=self.device).\
            repeat_interleave(Ks_per_bucket).expand((batch_size, -1))
        K_indices = K_indices.reshape([batch_size, padded_set_size])
        bucket_assignment = torch.zeros((batch_size, padded_set_size), dtype=torch.int32,
                                device=self.device).scatter_(1, K_indices, buckets)
        bucket_assignment = bucket_assignment[:, :set_size]
            
        return bucket_assignment
    
class Tree_PlusPlusQ_Builder(Tree_PlusPlus_Builder):
    
    def __init__(self, depth:int, dimension:int, iterations:int = 5,
                 noise_coeff:float = 0.5):
        super(Tree_PlusPlusQ_Builder, self).__init__(depth, dimension, iterations,
                                                     noise_coeff)
        
        self.projection_vectors = torch.tensor([], device=self.device)
        
    def save_projection_vectors(self, projection_vectors, level):
        
        vectors_per_batch = torch.pow(2, level).type(torch.int32)
        projection_vectors = projection_vectors.reshape((-1, vectors_per_batch,
                                                         self.dimension))
        self.projection_vectors = torch.cat((self.projection_vectors, projection_vectors),
                                       dim=1)
        
    def reset_projection_vectors(self):
        
        self.projection_vectors = torch.tensor([], device=self.device)
        
    def forward_Q(self, Q, retain_projection_vectors=False):
        
        batch_size = Q.shape[0]
        set_size = Q.shape[1]
        batch_inds = torch.arange(batch_size, device=self.device).repeat_interleave(
                                            set_size).reshape((batch_size, set_size))
        
        Q_assignments = torch.zeros((batch_size, set_size), device=self.device,
                                    dtype=torch.long)
        
        lower_index = 0
        higher_index = 1
        for level in torch.arange(self.depth, device=self.device):
            
            projection_vectors = self.projection_vectors[:,lower_index:higher_index,:]
            projection_vectors = projection_vectors[batch_inds, Q_assignments]
            projection_results = torch.mul(Q, projection_vectors).sum(-1)
            projection_results = (projection_results > 0).type(torch.int32)
            del projection_vectors
                
            Q_assignments = torch.mul(Q_assignments, 2)
            Q_assignments = torch.add(Q_assignments, projection_results)
            del projection_results
            
            lower_index = higher_index
            higher_index = higher_index*2 + 1
        
        # reset projection vectors for next step tree building
        if not retain_projection_vectors:
            self.reset_projection_vectors()
        
        return Q_assignments
        
    def split(self, K, K_indices, level):
        
        batch_size = K.shape[0]
        set_size = K.shape[1]
        noise_size_tensor = torch.tensor([batch_size, set_size], dtype=torch.int32,
                                         device=device)
        
        new_set_size = torch.floor(torch.div(set_size, 2)).int()
        new_batch_size = torch.mul(batch_size, 2).int()
        batch_inds = torch.arange(batch_size, device=self.device).repeat_interleave(
                                        set_size).reshape([batch_size, set_size])
        
        projection_vectors = self.initialize_projection_vectors(K)
        
        # through several iterations find a close to optimal projection direction
        for iteration in range(self.iterations):
            projection_results = torch.mul(K, projection_vectors).sum(dim=-1)
            rand_noise = self.noise_distribution.sample(noise_size_tensor).squeeze(-1)
            projection_results = projection_results + rand_noise
            del rand_noise
            sorted_inner_products, sort_indices = torch.sort(projection_results, dim=-1)
            del projection_results, sorted_inner_products, projection_vectors
            
            first_indices = sort_indices[:, :new_set_size]
            second_indices = sort_indices[:, new_set_size:]
            first_set = K[batch_inds[:, :new_set_size], first_indices]
            second_set = K[batch_inds[:, new_set_size:], second_indices]
            first_centers = torch.mean(first_set, dim=-2)
            second_centers = torch.mean(second_set, dim=-2)
            del first_indices, second_indices, first_set, second_set
            
            projection_vectors = torch.sub(second_centers, first_centers)
            projection_vectors_norm = torch.linalg.norm(projection_vectors, dim=-1).unsqueeze(-1)
            projection_vectors = torch.div(projection_vectors, projection_vectors_norm)
            projection_vectors = projection_vectors.unsqueeze(-2)
            del first_centers, second_centers
            
        # project each point into 1D and sort
        projection_results = torch.mul(K, projection_vectors).sum(dim=-1)
        sorted_inner_products, sort_indices = torch.sort(projection_results, dim=-1)
        # save splitting directions to forward queries
        first_centers = torch.mean(K[batch_inds[:, :new_set_size],
                                     sort_indices[:, :new_set_size]], dim=-2)
        second_centers = torch.mean(K[batch_inds[:, new_set_size:],
                                     sort_indices[:, new_set_size:]], dim=-2)
        projection_vectors = torch.sub(second_centers, first_centers)
        projection_vectors_norm = torch.linalg.norm(projection_vectors, dim=-1).unsqueeze(-1)
        projection_vectors = torch.div(projection_vectors, projection_vectors_norm)
        projection_vectors = projection_vectors.unsqueeze(-2)
        self.save_projection_vectors(projection_vectors, level)
        del projection_results, sorted_inner_products, projection_vectors, first_centers,\
            second_centers, projection_vectors_norm
        
        K = K[batch_inds, sort_indices]
        K = K.reshape([new_batch_size, new_set_size, K.shape[2]])
        K_indices = K_indices[batch_inds, sort_indices]
        K_indices = K_indices.reshape([new_batch_size, new_set_size])
        
        return K, K_indices
    

class Simple_Tree_Builder(nn.Module):
    
    def __init__(self, depth:int, dimension:int, max_split:bool = True):
        super(Simple_Tree_Builder, self).__init__()
        
        self.depth = depth
        self.n_leafs = int(math.pow(2, depth))
        self.device = device
        self.dimension = dimension
        self.max_split = max_split
    
    def max_direction(self, K):
        # randomly select one point
        batch_size = K.shape[0]
        set_size = K.shape[1]
        chosen_inds = torch.randint(0, set_size, [batch_size,], dtype=torch.long)
        batch_tensor = torch.arange(0, batch_size, 1, device=self.device, dtype=torch.long)
        chosen_points = K[batch_tensor, chosen_inds]
        #now according to distance choose first points
        distances = torch.sub(K, chosen_points.unsqueeze(-2))
        del chosen_inds, chosen_points
        distances = torch.linalg.norm(distances, dim=-1)
        distance_inds = torch.argmax(distances, dim=-1)
        first_points = K[batch_tensor, distance_inds]
        #now choose second points according to distance to first points
        distances = torch.sub(K, first_points.unsqueeze(-2))
        distances = torch.linalg.norm(distances, dim=-1)
        distance_inds = torch.argmax(distances, dim=-1)
        second_points = K[batch_tensor, distance_inds]
        del distances, distance_inds, batch_tensor
        #Finally build projection vector according to the chosen two points
        projection_vectors = torch.sub(first_points, second_points).unsqueeze(-2)
        
        return projection_vectors
    
    def rand_projection_vectors(self, K):
        # randomly select two different points
        batch_size = K.shape[0]
        set_size = K.shape[1]
        first_chosen_inds = torch.randint(0, set_size, [batch_size,], dtype=torch.long)
        second_chosen_inds = torch.randint(0, set_size-1, [batch_size,], dtype=torch.long)
        second_chosen_inds = second_chosen_inds + (second_chosen_inds == first_chosen_inds)
        batch_tensor = torch.arange(0, batch_size, 1, device=self.device, dtype=torch.long)
        first_points = K[batch_tensor, first_chosen_inds]
        second_points = K[batch_tensor, second_chosen_inds]
        #Finally build projection vector according to the chosen two points
        projection_vectors = torch.sub(first_points, second_points).unsqueeze(-2)
        
        return projection_vectors
    
    def split(self, K, K_indices):
        
        if self.max_split:
            projection_vectors = self.max_direction(K)
        else:
            projection_vectors = self.rand_projection_vectors(K)
        
        # project each point into 1D and sort
        projection_results = torch.mul(K, projection_vectors).sum(dim=-1)
        sorted_inner_products, sort_indices = torch.sort(projection_results, dim=-1)
        del projection_results, sorted_inner_products, projection_vectors
        
        new_set_size = torch.floor(torch.div(K.shape[1], 2))
        new_batch_size = torch.mul(K.shape[0], 2)
        
        batch_inds = torch.arange(K.shape[0], device=self.device).\
            repeat_interleave(K.shape[1]).reshape(
                [K.shape[0], K.shape[1]])
        K = K[batch_inds, sort_indices]
        K = K.reshape([new_batch_size, new_set_size, K.shape[2]])
        K_indices = K_indices[batch_inds, sort_indices]
        K_indices = K_indices.reshape([new_batch_size, new_set_size])
        
        return K, K_indices
        
    def forward(self, K, batch_size:int, set_size:int):
        
        # firstly pad the Key vectors such that the modified K can be evenly
        # distributed among the tree leafs
        n_padding_vectors = torch.remainder(-set_size, self.n_leafs)
        Ks_per_bucket = torch.ceil(torch.div(set_size, self.n_leafs))
        padded_set_size = torch.mul(Ks_per_bucket, self.n_leafs)
        # duplicate a random selection of key vectors to obtain padding vectors
        padding_vector_indices = torch.randint(0, set_size, [n_padding_vectors],
                                               dtype=torch.long, device=self.device)
        padding_vectors = K[:, padding_vector_indices]
        K = torch.cat((K, padding_vectors), dim=1)
        del padding_vectors, n_padding_vectors, padding_vector_indices
        K_indices = torch.arange(padded_set_size, device=self.device).repeat(
                                batch_size).reshape((batch_size, padded_set_size))
        
        # assign key vectors to tree leafs
        for level in torch.arange(self.depth):
            K, K_indices = self.split(K, K_indices)
        del K
            
        # find the leaf indices according to the reordered K_indices
        buckets = torch.arange(self.n_leafs, dtype=torch.int32, device=self.device).\
            repeat_interleave(Ks_per_bucket).expand((batch_size, -1))
        K_indices = K_indices.reshape([batch_size, padded_set_size])
        bucket_assignment = torch.zeros((batch_size, padded_set_size), dtype=torch.int32,
                                device=self.device).scatter_(1, K_indices, buckets)
        bucket_assignment = bucket_assignment[:, :set_size]
            
        return bucket_assignment
    
# Wrapper for the torch_sparse.spspmm(A, B) function which supports autograd
class Inner_Product_Wrapper(Function):

    @staticmethod
    def forward(ctx, a_indices, a_vectors, b_indices, b_vectors):
        
        output_vector = torch.mul(a_vectors[a_indices], b_vectors[b_indices]).sum(-1)
        
        ctx.save_for_backward(a_indices, a_vectors, b_indices, b_vectors)
        
        return output_vector

    @staticmethod
    def backward(ctx, grad_output_vector):
        a_indices, a_vectors, b_indices, b_vectors = ctx.saved_tensors
        
        grad_a_indices, grad_a_vectors, grad_b_indices, grad_b_vectors = None, None, None, None
        
        # compute the gradient with respect to the first vector collection
        grad_a_vectors = torch.mul(grad_output_vector.unsqueeze(-1), b_vectors[b_indices])
        grad_a_vectors = torch.zeros(a_vectors.shape, device=device).scatter_add_(
            0, a_indices.unsqueeze(-1).expand((-1, a_vectors.shape[1])), grad_a_vectors)
        
        # compute the gradient with respect to the second vector collection
        grad_b_vectors = torch.mul(grad_output_vector.unsqueeze(-1), a_vectors[a_indices])
        grad_b_vectors = torch.zeros(b_vectors.shape, device=device).scatter_add_(
            0, b_indices.unsqueeze(-1).expand((-1, b_vectors.shape[1])), grad_b_vectors)
        
        return grad_a_indices, grad_a_vectors, grad_b_indices, grad_b_vectors

class Simple_Attention_Tree(nn.Module):

    def __init__(self, n_leafs, dimension, scaled_softmax=True, max_split=True,
                 tree='simple', tree_plusplus_iterations=5,
                 tree_plusplus_noise_coefficient=0.2, att_softmax=True):
        super(Simple_Attention_Tree, self).__init__()
        
        # if the number of leafs isn't an integer exponential of 2, then round it
        # to the next higher integer exponential of 2
        self.depth = math.ceil(math.log2(n_leafs))
        self.n_leafs = int(2**self.depth)
        self.dimension = dimension
        self.max_split = max_split
        self.tree_plusplus_iterations = tree_plusplus_iterations
        self.tree_plusplus_noise_coefficient = tree_plusplus_noise_coefficient
        print('simple tree depth:', self.depth)
        print('number of leafs:', self.n_leafs)
        
        if tree == 'simple':
            self.tree_builder = torch.jit.script(Simple_Tree_Builder(self.depth,
                                            self.dimension, max_split=self.max_split))
            print('Tree Algorithm: simple')
            print('Max Split: ' + str(self.max_split))
        elif tree == 'plusplus':
            self.tree_builder = Tree_PlusPlus_Builder(self.depth, self.dimension,
                            iterations=self.tree_plusplus_iterations,
                            noise_coeff=self.tree_plusplus_noise_coefficient)
            print('Tree Algorithm: plusplus')
        elif tree == 'plusplusQ':
            self.tree_builder = Tree_PlusPlusQ_Builder(self.depth, self.dimension,
                            iterations=self.tree_plusplus_iterations,
                            noise_coeff=self.tree_plusplus_noise_coefficient)
            print('Tree Algorithm: plusplusQ')
        else:
            print('Illegal tree building method specified! Specify simple, plusplus'+\
                  ' or pusplusQ')
            print('Choosing simple tree construction therefore.')
            self.tree_builder = torch.jit.script(Simple_Tree_Builder(self.depth,
                                                self.dimension, max_split=self.max_split))
            tree = 'simple'
        self.tree = tree
        
        # parameters for attention normalization instead of softmax
        self.att_softmax = att_softmax
        self.alpha = nn.Parameter(torch.Tensor([1]), requires_grad=True)
        self.beta = nn.Parameter(torch.Tensor([0]), requires_grad=True)
            
        self.inner_product = Inner_Product_Wrapper.apply
        
        # epsilon which prevents division by 0
        self.epsilon = nn.Parameter(torch.Tensor([1e-32]), requires_grad=False)
        
        # create matrix which computes the coefficient sum for each tree node
        leaf_coefficient_matrix = torch.zeros((self.n_leafs - 1, self.n_leafs))
        for level in range(self.depth):
            split = int(math.pow(2, level))
            for node in range(int(math.pow(2, level))):
                row = int(math.pow(2, level) - 1) + node
                positive_start_column = int(self.n_leafs/split*node)
                negative_start_column = positive_start_column + int(0.5*self.n_leafs/split)
                end_column = positive_start_column + int(self.n_leafs/split)
                leaf_coefficient_matrix[row,
                    positive_start_column:negative_start_column] = 1
                leaf_coefficient_matrix[row,
                    negative_start_column:end_column] = -1
                
        self.leaf_coefficient_matrix = nn.Parameter(leaf_coefficient_matrix,
                                                    requires_grad=False)
        
        # constant which prevents division by zero
        self.one = nn.Parameter(torch.tensor([1]), requires_grad=False)
        
        # create matrix which computes the key and value sum for each tree node in upward pass
        upward_masking_matrix = torch.zeros((2*self.n_leafs - 2, self.n_leafs))
        for level in range(1, self.depth + 1):
            split = int(math.pow(2, level))
            for node in range(int(math.pow(2, level))):
                row = int(math.pow(2, level) - 2) + node
                start_column = int(self.n_leafs/split*node)
                end_column = start_column + int(self.n_leafs/split)
                upward_masking_matrix[row, start_column:end_column] = 1
        
        self.upward_masking_matrix = nn.Parameter(upward_masking_matrix,
                                                  requires_grad=False)        
        
        # create masking matrix for tree downward pass
        downward_masking_matrix = torch.zeros((2*self.n_leafs - 2, self.n_leafs))
        for level in range(1, self.depth + 1):
            split = int(math.pow(2, level))
            for node_pair in range(1, int(split/2) + 1):
                for node in range(2):
                    row = int(math.pow(2, level) - 2) + int(2*node_pair - 1) - node
                    start_column = int(self.n_leafs/split*(2*(node_pair-1) + node))
                    end_column = start_column + int(self.n_leafs/split)
                    downward_masking_matrix[row, start_column:end_column] = 1
        
        downward_indices = torch.zeros((self.n_leafs, self.depth))
        i = 0
        for row in downward_masking_matrix.transpose(-1,-2):
            nonzero_inds = (row==1).nonzero().squeeze()
            downward_indices[i] = nonzero_inds
            i += 1
            
        self.downward_indices = nn.Parameter(downward_indices.type(torch.long),
                                             requires_grad=False)
        
        if scaled_softmax:
            self.softmax_normalization_factor = nn.Parameter(torch.tensor(
                math.sqrt(dimension)), requires_grad=False)
        else:
            self.softmax_normalization_factor = nn.Parameter(torch.tensor(1),
                                                             requires_grad=False)

    def forward(self, K, Q, V):
        
        # assign each key to exactly one leaf in the tree
        # if do_timing:
        #     very_start = time()
        #     start = time()
        with torch.no_grad():
            batch_size = K.shape[0]
            set_size = K.shape[1]
            K_assignments = self.tree_builder(K, batch_size, set_size)
        # if do_timing:
            # torch.cuda.current_stream().synchronize()
            # print('total tree building time: ' + str(time() - start))
            # start = time()
        attention = self.calc_attention(K_assignments, Q, K, V)
        # if do_timing:
        #     torch.cuda.current_stream().synchronize()
        #     print('total attention calculation time: ' + str(time() - start))
        #     print('total simple tree attention time: ' + str(time() - very_start))
        return attention
    
    def upward(self, content_matrix, assigned_buckets, n_sets, dim,
               n_points, node_KNs):
        #create sparse block matrix for all content
        with torch.no_grad():
            right_shifts = torch.mul(assigned_buckets, dim).repeat_interleave(dim)
            block_horizontal_indices = torch.add(torch.arange(dim,
                            device=device).repeat(n_points*n_sets), right_shifts)
            block_n = dim*self.n_leafs*n_sets
        del right_shifts
        block_values = content_matrix.ravel()
        
        # calculate aggregated content of the tree nodes
        leaf_content = torch.zeros(block_n, device=device).scatter_add_(0,
            block_horizontal_indices, block_values)
        leaf_content = leaf_content.reshape([n_sets, self.n_leafs, dim])
        del block_horizontal_indices, block_values, block_n
        
        # calculate aggregated content of the tree nodes
        node_content = torch.matmul(self.upward_masking_matrix, leaf_content)
        del leaf_content
        
        # compute content average
        node_content = torch.div(node_content, torch.maximum(node_KNs, self.one))
        
        return node_content
    
    def compute_direct_attention_coefficients(self, K, Q, K_assignments, leaf_KNs,
                                              Q_assignments, n_sets, n_Qs):
        
        # compute direct attention coefficients:
        # create 2D very long matrix where each Query is repeated as many times as
        # there are Keys and Values in the leaf this Query is assigned to
        leaf_Q_matrix = torch.flatten(Q, start_dim=0, end_dim=1)
        with torch.no_grad():
            Q_indices = torch.arange(n_Qs*n_sets, dtype=torch.long, device=device
                                     ).repeat_interleave(leaf_KNs[Q_assignments])        
            # create 2D very long matrix where for each Query the whole set of keys within
            # the same leaf as that query is appended
            sorted_K_assignments, K_sort_indices = torch.sort(K_assignments, dim=-1)
            del sorted_K_assignments
            max_leaf_KN = torch.max(leaf_KNs)
            chosen_K_inds = torch.arange(max_leaf_KN, dtype=torch.long, device=device).\
                repeat(self.n_leafs*n_sets).reshape([self.n_leafs*n_sets, max_leaf_KN])
            del max_leaf_KN
            chosen_K_inds_mask = torch.sub(leaf_KNs.unsqueeze(-1), chosen_K_inds) > 0
            chosen_K_inds[1:] = torch.add(chosen_K_inds[1:], torch.cumsum(leaf_KNs[:-1],
                                                                    0).unsqueeze(-1))
            chosen_K_inds = chosen_K_inds[Q_assignments].ravel()
            chosen_K_inds_mask = chosen_K_inds_mask[Q_assignments].ravel()
            chosen_K_inds = chosen_K_inds[torch.nonzero(chosen_K_inds_mask).squeeze()]
            del chosen_K_inds_mask
            K_indices = K_sort_indices[chosen_K_inds]  
            del chosen_K_inds, K_sort_indices
        leaf_K_matrix = torch.flatten(K, start_dim=0, end_dim=1)
        
        # compute sparse attention matrix as using elment-wise matrix products
        QK_inner_products_matrix = self.inner_product(Q_indices, leaf_Q_matrix,
                                                      K_indices, leaf_K_matrix)
        del Q_indices, leaf_K_matrix, leaf_Q_matrix
        
        return QK_inner_products_matrix, K_indices
    
    def compute_aggregated_attention_coefficients(self, avg_node_Ks, Q, Q_assignments,
                                                  n_sets, n_Qs):
        
        # compute aggregated attention coefficients:
        # create 2D very long matrix where each Query is repeated as many times as
        # there are Keys and Values in the leaf this Query is assigned to
        leaf_Q_matrix = torch.flatten(Q, start_dim=0, end_dim=1)
        with torch.no_grad():
            Q_indices = torch.arange(n_Qs*n_sets, dtype=torch.long, device=device
                                    ).repeat_interleave(self.depth)
            # create 2D very long matrix where for each Query the whole set of keys within
            # the same leaf as that query is appended
            downward_indices = self.downward_indices.repeat((n_sets, 1, 1))
            batch_levels = torch.arange(0, n_sets*(2*self.n_leafs - 2),
                                2*self.n_leafs - 2, device=device)
            downward_indices = torch.add(downward_indices, batch_levels.unsqueeze(
                -1).unsqueeze(-1))
            del batch_levels
            downward_indices = downward_indices.flatten(start_dim=0, end_dim=1)
            K_indices = downward_indices[Q_assignments].ravel()
            del downward_indices
        avg_K_matrix = torch.flatten(avg_node_Ks, start_dim=0, end_dim=1)
        
        # compute sparse attention matrix as using elment-wise matrix products
        aggregated_QK_inner_products_matrix = self.inner_product(Q_indices, leaf_Q_matrix,
                                                                 K_indices, avg_K_matrix)
        del Q_indices, avg_K_matrix, leaf_Q_matrix
        
        return aggregated_QK_inner_products_matrix, K_indices
    
    def compute_weighted_softmax(self, QK_inner_products_matrix,
        aggregated_QK_inner_products_matrix, node_KNs, K_inds_aggregated,
        Q_assignments, leaf_KNs, n_sets, n_Qs):
        
        # firstly divide coefficients by softmax scaling factor in order
        # to compensate for different dimensions
        QK_inner_products_matrix = torch.div(QK_inner_products_matrix,
                                             self.softmax_normalization_factor)
        aggregated_QK_inner_products_matrix = torch.div(
            aggregated_QK_inner_products_matrix, self.softmax_normalization_factor)
        
        # exponentiate attention coefficients in a numerically stable way
        with torch.no_grad():
            max_val_direct = torch.max(QK_inner_products_matrix)
            max_val_aggregated = torch.max(aggregated_QK_inner_products_matrix)
            max_val = torch.maximum(max_val_direct, max_val_aggregated)
            max_val = torch.sub(max_val, 5)
        
        QK_inner_products_matrix = torch.sub(QK_inner_products_matrix, max_val)
        aggregated_QK_inner_products_matrix = torch.sub(
            aggregated_QK_inner_products_matrix, max_val)
        del max_val
        QK_inner_products_matrix = torch.exp(QK_inner_products_matrix)
        aggregated_QK_inner_products_matrix = torch.exp(
            aggregated_QK_inner_products_matrix)
        
        # weight aggregated inner products
        aggregated_QK_inner_products_matrix = torch.mul(
            aggregated_QK_inner_products_matrix, node_KNs.ravel()[K_inds_aggregated])
        
        # compute normalization coefficients
        aggregated_normalization_coefficients = aggregated_QK_inner_products_matrix.\
            reshape([n_sets*n_Qs, self.depth]).sum(-1)
        direct_normalization_coefficients = torch.zeros((n_sets*n_Qs), device=device).\
            scatter_add_(0, torch.repeat_interleave(torch.arange(n_sets*n_Qs,
                device=device), leaf_KNs[Q_assignments]), QK_inner_products_matrix)
        normalization_coefficients = torch.add(direct_normalization_coefficients,
                                               aggregated_normalization_coefficients)
        del direct_normalization_coefficients, aggregated_normalization_coefficients
        
        # normalize attention coefficients to obtain probabilities
        QK_inner_products_matrix = torch.div(QK_inner_products_matrix,
            torch.repeat_interleave(torch.max(normalization_coefficients, self.epsilon),
                                    leaf_KNs[Q_assignments]))
        aggregated_QK_inner_products_matrix = torch.div(
            aggregated_QK_inner_products_matrix, torch.repeat_interleave(
                torch.max(normalization_coefficients, self.epsilon), self.depth))
        del normalization_coefficients
        
        return QK_inner_products_matrix, aggregated_QK_inner_products_matrix
    
    def normalize_attention(self, QK_inner_products_matrix,
        aggregated_QK_inner_products_matrix, node_KNs, K_inds_aggregated,
        Q_assignments, leaf_KNs, n_sets, n_Qs):
        
        means = torch.add(torch.mul(aggregated_QK_inner_products_matrix,
            node_KNs.ravel()[K_inds_aggregated]).reshape((n_sets*n_Qs, self.depth)).sum(-1),
                        torch.zeros((n_sets*n_Qs), device=device).scatter_add_(0,
                        torch.repeat_interleave(torch.arange(n_sets*n_Qs, device=device),
                        leaf_KNs[Q_assignments]), QK_inner_products_matrix))
                                                                               
        means = torch.div(means, n_Qs)
        
        aggregated_QK_inner_products_matrix = torch.sub(aggregated_QK_inner_products_matrix,
                                                torch.repeat_interleave(means, self.depth))
        QK_inner_products_matrix = torch.sub(QK_inner_products_matrix,
                            torch.repeat_interleave(means, leaf_KNs[Q_assignments]))
        
        std_devs = torch.add(torch.mul(torch.pow(aggregated_QK_inner_products_matrix, 2),
            node_KNs.ravel()[K_inds_aggregated]).reshape((n_sets*n_Qs, self.depth)).sum(-1),
                torch.zeros((n_sets*n_Qs), device=device).scatter_add_(0,
                torch.repeat_interleave(torch.arange(n_sets*n_Qs, device=device),
                leaf_KNs[Q_assignments]), torch.pow(QK_inner_products_matrix, 2)))
        std_devs = torch.sqrt(torch.max(torch.div(std_devs, n_Qs), self.epsilon))
        
        aggregated_QK_inner_products_matrix = torch.div(aggregated_QK_inner_products_matrix,
                                        torch.repeat_interleave(std_devs, self.depth))
        QK_inner_products_matrix = torch.div(QK_inner_products_matrix,
                            torch.repeat_interleave(std_devs, leaf_KNs[Q_assignments]))
        
        aggregated_QK_inner_products_matrix = torch.mul(aggregated_QK_inner_products_matrix,
                                                        self.alpha)
        QK_inner_products_matrix = torch.mul(QK_inner_products_matrix, self.alpha)
        
        aggregated_QK_inner_products_matrix = torch.add(aggregated_QK_inner_products_matrix,
                                                        self.beta)
        QK_inner_products_matrix = torch.add(QK_inner_products_matrix, self.beta)
        
        return QK_inner_products_matrix, aggregated_QK_inner_products_matrix
    
    def compute_direct_weighted_sum(self, QK_inner_products_matrix, V, K_inds_direct,
        leaf_KNs, Q_assignments, n_sets, n_Qs, V_dim):
        
        V = torch.flatten(V, start_dim=0, end_dim=1)
        direct_weighted_sum_matrix = torch.mul(V[K_inds_direct],
                                               QK_inner_products_matrix.unsqueeze(-1))
        
        direct_weighted_sum_matrix = torch.zeros((n_sets*n_Qs, V_dim), device=device).\
            scatter_add_(0, torch.repeat_interleave(torch.arange(n_sets*n_Qs,
                device=device), leaf_KNs[Q_assignments]).unsqueeze(-1).expand((-1,
                    V_dim)), direct_weighted_sum_matrix)
        
        return direct_weighted_sum_matrix
    
    def compute_aggregated_weighted_sum(self, aggregated_QK_inner_products_matrix,
        avg_node_Vs, K_inds_aggregated, Q_assignments, n_sets, n_Qs, V_dim):
        
        avg_node_Vs = torch.flatten(avg_node_Vs, start_dim=0, end_dim=1)
        aggregated_weighted_sum_matrix = torch.mul(avg_node_Vs[K_inds_aggregated],
                                    aggregated_QK_inner_products_matrix.unsqueeze(-1))
        
        aggregated_weighted_sum_matrix = aggregated_weighted_sum_matrix.reshape(
            [n_sets*n_Qs, self.depth, V_dim]).sum(1)
        
        return aggregated_weighted_sum_matrix
    
    def calc_attention(self, K_assignments, Q, K, V):
        
        n_Ks = K.shape[-2]
        n_Qs = Q.shape[-2]
        n_Vs = V.shape[-2]
        K_dim = K.shape[-1]
        Q_dim = Q.shape[-1]
        V_dim = V.shape[-1]
        n_sets = K.shape[0]
    
        # calculate number of Keys assigned to each node
        with torch.no_grad():
            batch_levels = torch.arange(0, n_sets*self.n_leafs,
                                self.n_leafs, device=device).repeat_interleave(n_Ks)
            # sorted_K_assignments, K_sort_indices = torch.sort(K_assignments, dim=-1)
            K_assignments = torch.add(K_assignments.ravel(), batch_levels)
            # sorted_K_assignments = torch.add(sorted_K_assignments.ravel(), batch_levels)
            leaf_KNs = torch.bincount(K_assignments, minlength=self.n_leafs*n_sets)
            node_KNs = torch.matmul(self.upward_masking_matrix,
                        leaf_KNs.reshape([n_sets, self.n_leafs, 1]).type(torch.float))
            # del batch_levels
            
        # compute key and value averages in all nodes
        avg_node_Ks = self.upward(K, K_assignments, n_sets, K_dim, n_Ks, node_KNs)
        avg_node_Vs = self.upward(V, K_assignments, n_sets, V_dim, n_Vs, node_KNs)
        # end upward pass
        
        # begin downward pass
        # downward pass of queries
        # the queries are assigned to the leafs where they have largest inner products
        with torch.no_grad():
            if self.tree == 'plusplusQ':
                Q_assignments = self.tree_builder.forward_Q(Q)
            else:
                Q_intersections = torch.matmul(Q,
                                    avg_node_Ks[:,-self.n_leafs:].transpose(-1,-2))
                Q_assignments = torch.argmax(Q_intersections, dim=-1)
                del Q_intersections
            batch_levels = torch.arange(0, n_sets*self.n_leafs,
                                self.n_leafs, device=device).repeat_interleave(n_Qs)
            Q_assignments = torch.add(Q_assignments.ravel(), batch_levels)
            del batch_levels
        
        # compute direct attention coefficients
        QK_inner_products_matrix, K_inds_direct = self.compute_direct_attention_coefficients(
            K, Q, K_assignments, leaf_KNs, Q_assignments, n_sets, n_Qs)
        del K, K_assignments
        
        # compute aggregated attention coefficients
        aggregated_QK_inner_products_matrix, K_inds_aggregated =\
            self.compute_aggregated_attention_coefficients(avg_node_Ks, Q, Q_assignments,
                                                           n_sets, n_Qs)
        del avg_node_Ks, Q
            
        # compute softmax for each query
        if self.att_softmax:
            QK_inner_products_matrix, aggregated_QK_inner_products_matrix =\
                self.compute_weighted_softmax(QK_inner_products_matrix,
                    aggregated_QK_inner_products_matrix, node_KNs, K_inds_aggregated,
                    Q_assignments, leaf_KNs, n_sets, n_Qs)
        else:
            QK_inner_products_matrix, aggregated_QK_inner_products_matrix =\
                self.normalize_attention(QK_inner_products_matrix,
                    aggregated_QK_inner_products_matrix, node_KNs, K_inds_aggregated,
                    Q_assignments, leaf_KNs, n_sets, n_Qs)
        del node_KNs
            
        # compute direct weighted sum of attention values
        direct_weighted_sum_matrix = self.compute_direct_weighted_sum(
            QK_inner_products_matrix, V, K_inds_direct, leaf_KNs, Q_assignments,
            n_sets, n_Qs, V_dim)
        del QK_inner_products_matrix, V, K_inds_direct, leaf_KNs
        
        # compute aggregated weighted sum of attention values
        aggregated_weighted_sum_matrix = self.compute_aggregated_weighted_sum(
            aggregated_QK_inner_products_matrix, avg_node_Vs, K_inds_aggregated,
            Q_assignments, n_sets, n_Qs, V_dim)
        del aggregated_QK_inner_products_matrix, avg_node_Vs, K_inds_aggregated,\
            Q_assignments
        
        attention_matrix = torch.add(direct_weighted_sum_matrix,
                                     aggregated_weighted_sum_matrix)
        del direct_weighted_sum_matrix, aggregated_weighted_sum_matrix
        attention_matrix = attention_matrix.reshape([n_sets, n_Qs, V_dim])
        
        return attention_matrix        
    
    def get_n_leafs(self):
        
        return self.n_leafs
    
    def get_depth(self):
        
        return self.depth

class MAB(nn.Module):
    
    def __init__(self, dim_Q, dim_K, dim_V, num_heads, ln=False):
        super(MAB, self).__init__()
        self.dim_V = dim_V
        self.num_heads = num_heads
        self.fc_q = nn.Linear(dim_Q, dim_V)
        self.fc_k = nn.Linear(dim_K, dim_V)
        self.fc_v = nn.Linear(dim_K, dim_V)
        if ln:
            self.ln0 = nn.LayerNorm(dim_V)
            self.ln1 = nn.LayerNorm(dim_V)
        self.fc_o = nn.Linear(dim_V, dim_V)

    def forward(self, Q, K):
        
        Q = self.fc_q(Q)
        K, V = self.fc_k(K), self.fc_v(K)

        dim_split = self.dim_V // self.num_heads
        Q_ = torch.cat(Q.split(dim_split, 2), 0)
        K_ = torch.cat(K.split(dim_split, 2), 0)
        V_ = torch.cat(V.split(dim_split, 2), 0)

        A = torch.softmax(Q_.bmm(K_.transpose(1,2))/math.sqrt(self.dim_V), 2)
        O = torch.cat((Q_ + A.bmm(V_)).split(Q.size(0), 0), 2)
        O = O if getattr(self, 'ln0', None) is None else self.ln0(O)
        O = O + F.relu(self.fc_o(O))
        O = O if getattr(self, 'ln1', None) is None else self.ln1(O)
        return O
    
class SAB(nn.Module):
    
    def __init__(self, dim_in, dim_out, num_heads, ln=False):
        super(SAB, self).__init__()
        self.mab = MAB(dim_in, dim_in, dim_out, num_heads, ln=ln)

    def forward(self, X):
        return self.mab(X, X)
    
class PMA(nn.Module):
    
    def __init__(self, dim, num_heads, num_seeds, ln=False):
        super(PMA, self).__init__()
        self.S = nn.Parameter(torch.Tensor(1, num_seeds, dim))
        nn.init.xavier_uniform_(self.S)
        self.mab = MAB(dim, dim, dim, num_heads, ln=ln)

    def forward(self, X):
        return self.mab(self.S.repeat(X.size(0), 1, 1), X)        
            
class Tree_MAB(nn.Module):
    
    def __init__(self, dim_Q, dim_K, dim_V, num_heads, num_leafs, ln=False,
                 max_tree_construction=True, tree='simple', att_softmax=True):
        super(Tree_MAB, self).__init__()
        self.dim_V = dim_V
        self.num_heads = num_heads
        self.fc_q = nn.Linear(dim_Q, dim_V)
        self.fc_k = nn.Linear(dim_K, dim_V)
        self.fc_v = nn.Linear(dim_K, dim_V)
        tree_dim = dim_V // num_heads
        self.att_tree = Simple_Attention_Tree(num_leafs, tree_dim,
                                              max_split=max_tree_construction,
                                              tree=tree, att_softmax=att_softmax)
        if ln:
            self.ln0 = nn.LayerNorm(dim_V)
            self.ln1 = nn.LayerNorm(dim_V)
        self.fc_o = nn.Linear(dim_V, dim_V)

    def forward(self, Q, K):
        
        Q = self.fc_q(Q)
        K, V = self.fc_k(K), self.fc_v(K)

        dim_split = self.dim_V // self.num_heads
        Q_ = torch.cat(Q.split(dim_split, 2), 0)
        K_ = torch.cat(K.split(dim_split, 2), 0)
        V_ = torch.cat(V.split(dim_split, 2), 0)

        O = self.att_tree(K_, Q_, V_)
        O = torch.add(Q_, O)
        O = torch.cat(O.split(Q.size(0), 0), 2)
        O = O if getattr(self, 'ln0', None) is None else self.ln0(O)
        O = O + F.relu(self.fc_o(O))
        O = O if getattr(self, 'ln1', None) is None else self.ln1(O)
        
        return O
    
class Mult_Tree_MAB(nn.Module):
    
    def __init__(self, dim_Q, dim_K, dim_V, num_heads, num_leafs, num_trees,
                 ln=False, max_tree_construction=True, tree='simple'):
        super(Mult_Tree_MAB, self).__init__()
        self.dim_V = dim_V
        self.num_heads = num_heads
        self.num_trees = num_trees
        self.fc_q = nn.Linear(dim_Q, dim_V)
        self.fc_k = nn.Linear(dim_K, dim_V)
        self.fc_v = nn.Linear(dim_K, dim_V)
        tree_dim = dim_V // num_heads
        self.att_tree = Simple_Attention_Tree(num_leafs, tree_dim,
                                              max_split=max_tree_construction,
                                              tree=tree)
        if ln:
            self.ln0 = nn.LayerNorm(dim_V)
            self.ln1 = nn.LayerNorm(dim_V)
        self.fc_o = nn.Linear(dim_V, dim_V)

    def forward(self, Q, K):
        
        Q = self.fc_q(Q)
        K, V = self.fc_k(K), self.fc_v(K)

        dim_split = self.dim_V // self.num_heads
        Q_ = torch.cat(Q.split(dim_split, 2), 0)
        K_ = torch.cat(K.split(dim_split, 2), 0)
        V_ = torch.cat(V.split(dim_split, 2), 0)
        orig_Q_size = Q_.size(0)
        
        Q_ = Q_.repeat((self.num_trees, 1, 1))
        K_ = K_.repeat((self.num_trees, 1, 1))
        V_ = V_.repeat((self.num_trees, 1, 1))

        O = self.att_tree(K_, Q_, V_)
        O = torch.add(Q_, O)
        
        O = torch.stack(O.split(orig_Q_size, 0), 0).sum(0)
        O = torch.div(O, self.num_trees)
        
        O = torch.cat(O.split(Q.size(0), 0), 2)
        O = O if getattr(self, 'ln0', None) is None else self.ln0(O)
        O = O + F.relu(self.fc_o(O))
        O = O if getattr(self, 'ln1', None) is None else self.ln1(O)
        
        return O

class Tree_SAB(nn.Module):
    
    def __init__(self, dim_in, dim_out, num_heads, num_leafs, ln=False,
                 max_tree_construction=True, tree='simple', num_trees=1,
                 att_softmax=True):
        super(Tree_SAB, self).__init__()
        if num_trees > 1:
            self.mab = Mult_Tree_MAB(dim_in, dim_in, dim_out, num_heads, num_leafs,
                                     num_trees, ln=ln,
                                     max_tree_construction=max_tree_construction, tree=tree)
        else:
            self.mab = Tree_MAB(dim_in, dim_in, dim_out, num_heads, num_leafs, ln=ln,
                                max_tree_construction=max_tree_construction, tree=tree,
                                att_softmax=att_softmax)

    def forward(self, X):
        return self.mab(X, X)
    
class Tree_ISAB(nn.Module):

    def __init__(self, dim_in, dim_out, num_heads, num_leafs, ln=False,
                 max_tree_construction=True, tree='simple', num_trees=1):
        super(Tree_ISAB, self).__init__()
        if num_trees > 1:
            self.sab = Tree_SAB(dim_in, dim_out, num_heads, num_leafs, ln=ln,
                                max_tree_construction=max_tree_construction,
                                tree=tree, num_trees=num_trees)
            self.mab = Mult_Tree_MAB(dim_in, dim_out, dim_out, num_heads, num_leafs,
                            num_trees, ln=ln, max_tree_construction=max_tree_construction,
                            tree=tree)
        else:
            self.sab = Tree_SAB(dim_in, dim_out, num_heads, num_leafs, ln=ln,
                                max_tree_construction=max_tree_construction, tree=tree)
            self.mab = Tree_MAB(dim_in, dim_out, dim_out, num_heads, num_leafs, ln=ln,
                                max_tree_construction=max_tree_construction, tree=tree)
    
    def forward(self, X):
        return self.mab(X, self.sab(X))
    
class Full_ISAB(nn.Module):

    def __init__(self, dim_in, dim_out, num_heads, ln=False):
        super(Full_ISAB, self).__init__()
        self.sab = SAB(dim_in, dim_out, num_heads, ln=ln)
        self.mab = MAB(dim_in, dim_out, dim_out, num_heads, ln=ln)
    
    def forward(self, X):
        H = self.sab(X)
        return self.mab(X, H)
    
class ISAB(nn.Module):
    def __init__(self, dim_in, dim_out, num_heads, num_inds, ln=False):
        super(ISAB, self).__init__()
        self.I = nn.Parameter(torch.Tensor(1, num_inds, dim_out))
        nn.init.xavier_uniform_(self.I)
        self.mab0 = MAB(dim_out, dim_in, dim_out, num_heads, ln=ln)
        self.mab1 = MAB(dim_in, dim_out, dim_out, num_heads, ln=ln)

    def forward(self, X):
        H = self.mab0(self.I.repeat(X.size(0), 1, 1), X)
        return self.mab1(X, H)
  
class reverse_ISAB(nn.Module):
    def __init__(self, dim_in, dim_out, num_heads, num_inds, ln=False):
        super(reverse_ISAB, self).__init__()
        self.I = nn.Parameter(torch.Tensor(1, num_inds, dim_out))
        nn.init.xavier_uniform_(self.I)
        self.mab0 = MAB(dim_out, dim_in, dim_out, num_heads, ln=ln)
        self.mab1 = MAB(dim_in, dim_out, dim_out, num_heads, ln=ln)

    def forward(self, X, Inp):
        H = self.mab0(self.I.repeat(X.size(0), 1, 1), X)
        return self.mab1(Inp, H)
    

# test functions from now on
def exact_tree_timing(tree, n, d):
    Q = torch.randn((4,n,d), requires_grad=True).to(device)
    K = torch.randn((4,k,d), requires_grad=True).to(device)
    V = torch.randn((4,k,d), requires_grad=True).to(device)
    
    # time measurement
    torch.cuda.current_stream().synchronize()
    start_time = time()
    attention = tree(Q, K, V)
    torch.cuda.current_stream().synchronize()
    forward_time = time() - start_time
    loss = torch.sum(attention)
    # print('loss: ', loss)
    torch.cuda.current_stream().synchronize()
    start_time = time()
    loss.backward()
    torch.cuda.current_stream().synchronize()
    backward_time = time() - start_time
    total_time = forward_time + backward_time
    # print('forward time: ', forward_time)
    # print('backward time: ', backward_time)
    # print('total time: ', total_time)
    return forward_time, total_time

def exact_dense_timing(n, d):
    Q = torch.randn((4,n,d), requires_grad=True).to(device)
    K = torch.randn((4,k,d), requires_grad=True).to(device)
    V = torch.randn((4,k,d), requires_grad=True).to(device)
    
    # time measurement
    torch.cuda.current_stream().synchronize()
    start_time = time()
    attention = torch.softmax(Q.bmm(K.transpose(1,2))/math.sqrt(d), 2).bmm(V)
    torch.cuda.current_stream().synchronize()
    forward_time = time() - start_time
    loss = torch.sum(attention)
    # print('loss: ', loss)
    torch.cuda.current_stream().synchronize()
    start_time = time()
    loss.backward()
    torch.cuda.current_stream().synchronize()
    backward_time = time() - start_time
    total_time = forward_time + backward_time
    # print('forward time: ', forward_time)
    # print('backward time: ', backward_time)
    # print('total time: ', total_time)
    return forward_time, total_time

def exact_tree_memory_measurement(tree, n, d):    
    torch.cuda.reset_peak_memory_stats(device=device)
    Q = torch.randn((4,n,d), requires_grad=True).to(device)
    K = torch.randn((4,k,d), requires_grad=True).to(device)
    V = torch.randn((4,k,d), requires_grad=True).to(device)
    
    # memory measurement
    attention = tree(K, Q, V)
    
    forward_memory = torch.cuda.max_memory_allocated(device=device)/2**20

    loss = torch.sum(attention)
    loss.backward()
    
    total_memory = torch.cuda.max_memory_allocated(device=device)/2**20
    print('forward:, ', forward_memory)
    print('total: ', total_memory)
    
    return forward_memory, total_memory

def exact_dense_memory_measurement(n, d):
    torch.cuda.reset_peak_memory_stats(device=device)
    Q = torch.randn((4,n,d), requires_grad=True).to(device)
    K = torch.randn((4,k,d), requires_grad=True).to(device)
    V = torch.randn((4,k,d), requires_grad=True).to(device)
    
    # memory measurement
    attention = torch.softmax(Q.bmm(K.transpose(1,2))/math.sqrt(d), 2).bmm(V)
    
    forward_memory = torch.cuda.max_memory_allocated(device=device)/2**20

    loss = torch.sum(attention)
    loss.backward()
    
    total_memory = torch.cuda.max_memory_allocated(device=device)/2**20
    print('forward:, ', forward_memory)
    print('total: ', total_memory)
    
    return forward_memory, total_memory
    
  
if __name__ == '__main__':
    
    # from torch.profiler import profile, ProfilerActivity
    import numpy as np
    import pickle
    
    set_sizes = [100, 500, 1000, 1500, 2000, 2500, 3000, 3500, 4000, 4500, 5000,
                 6000, 7000, 8000, 9000, 10000, 12000, 14000, 16000, 18000, 20000,
                 22000, 24000, 26000, 28000, 30000]
    dense_set_sizes = [100, 500, 1000, 1500, 2000, 2500, 3000, 3500, 4000, 4500, 5000,
                       6000, 7000, 8000, 9000, 10000, 12000]
    tree_kinds = ['random', 'max', 'plusplus', 'plusplusQ']
    
    d = 64
    
    # tree time measurement
    tree_dict = {}
    for tree_kind in tree_kinds:
        fw_time_usage = []
        tot_time_usage = []
        stats_dict = {}
        for n in set_sizes:
            k = n
            if tree_kind == 'random':
                n_leafs = int(math.sqrt(k))
                tree = Simple_Attention_Tree(n_leafs, d, tree='simple',
                                             max_split=False).to(device)
            elif tree_kind == 'max':
                n_leafs = int(math.sqrt(k))
                tree = Simple_Attention_Tree(n_leafs, d, tree='simple',
                                             max_split=True).to(device)
            elif tree_kind == 'plusplus':
                n_leafs = int(math.sqrt(k))
                tree = Simple_Attention_Tree(n_leafs, d, tree='plusplus').to(device)
            elif tree_kind == 'plusplusQ':
                n_leafs = int(2**math.floor(math.log2(n)))
                tree = Simple_Attention_Tree(n_leafs, d, tree='plusplusQ').to(device)
            else:
                print('not good')
                import sys
                sys.exit()
            fw_times = []
            tot_times = []
            for i in range(5):
                
                forward_time, total_time = exact_tree_timing(tree, n, d)
                fw_times.append(forward_time)
                tot_times.append(total_time)
                
            fw_time_usage.append(np.average(fw_times))
            tot_time_usage.append(np.average(tot_times))
            
        stats_dict.update({'fw_times': fw_time_usage})
        stats_dict.update({'tot_times': tot_time_usage})
        tree_dict.update({tree_kind: stats_dict})
        del tree
        
    # dense time measurement
    dense_dict = {}
    fw_time_usage = []
    tot_time_usage = []
    for n in dense_set_sizes:
        print('n: ', n)
        k = n
        fw_times = []
        tot_times = []
        for i in range(5):
            
            forward_time, total_time = exact_dense_timing(n, d)
            fw_times.append(forward_time)
            tot_times.append(total_time)
            
        fw_time_usage.append(np.average(fw_times))
        tot_time_usage.append(np.average(tot_times))
        
    dense_dict.update({'fw_times': fw_time_usage})
    dense_dict.update({'tot_times': tot_time_usage})
    
    # tree memory measurement    
    for tree_kind in tree_kinds:
        fw_memory_usage = []
        tot_memory_usage = []
        stats_dict = {}
        for n in set_sizes:
            k = n
            if tree_kind == 'random':
                n_leafs = int(math.sqrt(k))
                tree = Simple_Attention_Tree(n_leafs, d, tree='simple',
                                             max_split=False).to(device)
            elif tree_kind == 'max':
                n_leafs = int(math.sqrt(k))
                tree = Simple_Attention_Tree(n_leafs, d, tree='simple',
                                             max_split=True).to(device)
            elif tree_kind == 'plusplus':
                n_leafs = int(math.sqrt(k))
                tree = Simple_Attention_Tree(n_leafs, d, tree='plusplus').to(device)
            elif tree_kind == 'plusplusQ':
                n_leafs = int(2**math.floor(math.log2(n)))
                tree = Simple_Attention_Tree(n_leafs, d, tree='plusplusQ').to(device)
            else:
                print('not good')
                import sys
                sys.exit()
            fw_memory = []
            tot_memory = []
            for i in range(5):
                
                forward_memory, total_memory = exact_tree_memory_measurement(tree, n, d)
                fw_memory.append(forward_memory)
                tot_memory.append(total_memory)
                
            fw_memory_usage.append(np.average(fw_memory))
            tot_memory_usage.append(np.average(tot_memory))
            
        tree_dict[tree_kind].update({'fw_memory': fw_memory_usage})
        tree_dict[tree_kind].update({'tot_memory': tot_memory_usage})
        tree_dict[tree_kind].update({'set_sizes': set_sizes})
        del tree
        
        # save this runs data to file
        path = tree_kind + '_tree_performance.pkl'
        savefile = open(path, "wb")
        pickle.dump(tree_dict[tree_kind], savefile)
        savefile.close()
    
    # dense memory measurement
    fw_memory_usage = []
    tot_memory_usage = []
    for n in dense_set_sizes:
        print('n: ', n)
        k = n
        fw_memory = []
        tot_memory = []
        for i in range(5):
            
            forward_memory, total_memory = exact_dense_memory_measurement(n, d)
            fw_memory.append(forward_memory)
            tot_memory.append(total_memory)
            
        fw_memory_usage.append(np.average(fw_memory))
        tot_memory_usage.append(np.average(tot_memory))
        
    dense_dict.update({'fw_memory': fw_memory_usage})
    dense_dict.update({'tot_memory': tot_memory_usage})
    dense_dict.update({'set_sizes': dense_set_sizes})
    
    # save this runs data to file
    path = 'dense_performance.pkl'
    savefile = open(path, "wb")
    pickle.dump(dense_dict, savefile)
    savefile.close()
                
    
    n = 2000
    k = n

    Q = torch.randn((4,n,d), requires_grad=True).to(device)
    K = torch.randn((4,k,d), requires_grad=True).to(device)
    V = torch.randn((4,k,d), requires_grad=True).to(device)

    #tree = Simple_Attention_Tree(int(math.sqrt(k)), d, tree='plusplusQ').to(device)
    tree = Simple_Attention_Tree(int(2**math.floor(math.log2(n))), d,
                                 tree='plusplusQ').to(device)

    # time measurement
    start_time = time()
    attention = tree(Q, K, V)
    torch.cuda.current_stream().synchronize()
    forward_time = time() - start_time
    loss = torch.sum(attention)
    print('loss: ', loss)
    start_time = time()
    loss.backward()
    torch.cuda.current_stream().synchronize()
    backward_time = time() - start_time
    total_time = forward_time + backward_time
    print('forward time: ', forward_time)
    print('backward time: ', backward_time)
    print('total time: ', total_time)
    
    # Q_grad_new = torch.clone(Q.grad)
    # K_grad_new = torch.clone(K.grad)
    # V_grad_new = torch.clone(V.grad)
    # attention_new = torch.clone(attention)
    # # print(attention_new[0])
    
    # Q.grad = torch.zeros(Q.shape)
    # K.grad = torch.zeros(K.shape)
    # V.grad = torch.zeros(V.shape)
    
    # attention = tree_old(K, Q, V, assigned_buckets)
    # loss = torch.sum(attention)
    # # print(loss)
    # loss.backward()
    # Q_grad_old = torch.clone(Q.grad)
    # K_grad_old = torch.clone(K.grad)
    # V_grad_old = torch.clone(V.grad)
    # attention_old = torch.clone(attention)
    # # print(attention_old[0])
    
    # print(torch.sum(torch.abs(torch.sub(attention_new, attention_old))))
    # print(torch.sub(torch.sum(attention_new), torch.sum(attention_old)))
    # print(torch.sum(torch.abs(torch.sub(Q_grad_new, Q_grad_old))))
    # print(torch.sum(torch.abs(torch.sub(K_grad_new, K_grad_old))))
    # print(torch.sum(torch.abs(torch.sub(V_grad_new, V_grad_old))))
    
    # sys.exit()
    
    # transformer_network_parameters = [Q, K, V]
    # optimizer = torch.optim.Adam(transformer_network_parameters)
    start = time()
    attention = tree1(K, Q, V)
    loss = torch.sum(attention)
    print('forward time: ' + str(time() - start))
    # print('attention: x')
    # print(attention)
    start = time()
    loss.backward()
    print('backward time:' + str(time() - start))
    # print(attention)
    print(loss)
    # print('V grad')
    print(torch.max(Q.grad))
    print(torch.max(K.grad))
    print(torch.max(V.grad))
    
    attention2 = tree2(K, Q, V)
    loss2 = torch.sum(attention2)
    print(loss2)
    del Q, K, V