#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Oct 25 20:10:37 2021

@author: sandro
"""

import torch
import math
import argparse
import sys
import pickle

import numpy as np
import os
from time import time, sleep
from torch.utils import tensorboard
from omegaconf import OmegaConf
from datetime import datetime

from Point_Segmentation_Dataloader import ShapeNet_Dataloader
from Point_Segmentation_models import Tree_Segmentation_Network, IV_Segmentation_Network,\
    Full_Segmentation_Network, Enhanced_Tree_Segmentation_Network, Classification_Loss,\
        Accuracy_Loss, IoU_Loss, Standard_Transformer_Segmentation_Network,\
            Deep_Standard_Transformer_Segmentation_Network,\
                Shallow_Standard_Transformer_Segmentation_Network,\
                    Standard_Transformer_notree, IV_Segmentation_Network_deep
from Visualization.Visualization_Tools import mean_average_plot

device = "cuda" if torch.cuda.is_available() else "cpu"
print("Using {} device".format(device))

run_performances_folder = 'run_performances'
# be aware of the running time in order to avoid job cancellations due to time limit
job_launch_time = time()
time_margin = 600

# parse command line arguments
parser = argparse.ArgumentParser(description='Parameters for point cloud '+\
                                 'segmentation pipeline')
parser.add_argument('-model_batch_size', metavar='model_batch_size', type=np.uint16,
                    default=16, help='Model Batch size')
parser.add_argument('-benchmark_batch_size', metavar='benchmark_batch_size', type=np.uint16,
                    default=4, help='Benchmark Batch size')
parser.add_argument('-batch_size', metavar='batch_size', type=np.uint16, default=10,
                    help='Batch size')
parser.add_argument('-steps', metavar='N_steps', type=int, default=70000,
                    help='Number of steps to train the network')
parser.add_argument('-lr', metavar='learning_rate', type=float, default=1e-3,
                    help='Learning rate')
parser.add_argument('-runs', metavar='N_training_runs', type=np.uint8, default=5,
                    help='Number of training repetitions')
parser.add_argument('-logdir', metavar='tensorboard_logdir', type=str, default='',
                    help='Directory where tensorboard logs should be written to')
parser.add_argument('-log', metavar='write_tensorboard_log', type=bool, default=False,
                    help='Wheter to write a tensorboard log or not')
parser.add_argument('-timing', metavar='do_exact_timing', type=bool, default=False,
                    help='Wheter to perform exact timing of fw and bw pass')
parser.add_argument('-ln', metavar='do_layer_normalization', type=bool, default=False,
                    help='Wheter to perform layer normalization in the Set Transformer')
parser.add_argument('-tree_depth', metavar='att_tree_depth', type=int, default=6,
                    help='How deep the attention tree should be')
parser.add_argument('-benchmark', metavar='benchmark', type=str, default=None,
                    help='Wheter to use Full Set Transformer or Basic Set Transformer')
parser.add_argument('-model', metavar='model', type=str, default=None,
                    help='Which model with tree attention to use')
parser.add_argument('-lr_dec_int', metavar='lr_decrease_interval', type=int, default=10000,
                    help='After how many epochs to multiply the lr by factor 0.8')
parser.add_argument('-data_dir', metavar='data_directory', type=str,
                    default='/scratch/snx3000/sbaumgar/points_3d_data',
                    help='Where to store/load the dataset')
parser.add_argument('-model_label', metavar='model_label', type=str, default='model',
                    help='label of the model error curve in the plot')
parser.add_argument('-benchmark_label', metavar='benchmark_label', type=str,
                    default='benchmark',
                    help='label of the benchmark error curve in the plot')
parser.add_argument('-plot_file', metavar='plot_file', type=str,
                    default='point_cloud_segmentation_plot.png',
                    help='filename of the model vs. benchmark error plot')
parser.add_argument('-d_augmentation', metavar='data_augmentation', type=str,
                    default='Full',
                    help='Data augmentation measures for training. Must be \'Full\', '+\
                        '\'GaussianOnly\', \'RotationOnly\', \'ScalingOnly\' or \'No\'.')
parser.add_argument('-tree_algorithm', metavar='tree_algorithm',
                    type=str, default='simple',
                    help='which algorithm should be used to construct the binary tree. ' +\
                        'Must be \'simple\', \'max\' or \'plusplus\'.')
parser.add_argument('-jobid', metavar='jobid', type=str, default='0',
                    help='Job ID of the job on piz daint')
parser.add_argument('-mult_trees', metavar='mult_trees', type=int, default=1,
                    help='Use an integer greater 1 to specify the amount of trees' +\
                        ' to be used in tree ensembles')
parser.add_argument('-test_loss', metavar='test_loss', type=str, default='Accuracy',
                    help='Specify wheter use Accuracy or IoU to measure model performance')
parser.add_argument('-num_heads', metavar='num_heads', type=int, default=8,
                    help='Specify the number of attention heads')
parser.add_argument('-dim_hidden', metavar='dim_hidden', type=int, default=128,
                    help='Specify the dimension of the hidden layers')
parser.add_argument('-category', metavar='category', type=str, default='Airplane',
                    help='Specify the dataset category')
parser.add_argument('-warmup_steps', metavar='warmup_steps', type=int, default=1000,
                    help='number of linear warmup steps for training')

args = parser.parse_args()
model_batch_size = args.model_batch_size
benchmark_batch_size = args.benchmark_batch_size
n_steps = args.steps
lr = args.lr
n_training_runs = args.runs
tensorboard_logdir = args.logdir
write_log = args.log
exact_timing = args.timing
ln = args.ln
tree_depth = args.tree_depth
benchmark = args.benchmark
model_architecture = args.model
lr_decrease_interval = args.lr_dec_int
directory = args.data_dir
model_label = args.model_label
benchmark_label = args.benchmark_label
plot_file = args.plot_file
data_augmentation = args.d_augmentation
tree_algorithm = args.tree_algorithm
job_id = args.jobid
num_trees = args.mult_trees
test_loss = args.test_loss
num_heads = args.num_heads
dim_hidden = args.dim_hidden
category = args.category
warmup_steps = max(args.warmup_steps, 0)

if not (benchmark in ['Full', 'Standard_Transformer', None] or 'IV' in benchmark):
    print('illegal benchmark! Please use \'Full\', \'Standard_Transformer\' or \'IV\'')
    sys.exit()
    
if not (model_architecture in ['Set_Transformer', 'Enhanced_Set_Transformer',
                               'Set_Transformer_Extra_Deep', 'Standard_Transformer',
                               'Deep_Standard_Transformer', 'Shallow_Standard_Transformer',
                               None]):
    print('illegal model! Please use \'Set_Transformer\', \'Enhanced_Set_Transformer\''+\
          ', \'Set_Transformer_Extra_Deep\', \'(Deep_)Standard_Transformer\' or nothing')
    sys.exit()
    
if tree_depth < 1:
    tree_depth = 1
    
if data_augmentation == 'Full':
    data_augmentation_yaml = """
    RandomGaussianNoise:
        params:
            sigma: 0.01
            clip: 0.05
    RandomRotation:
        params:
            max_degrees: 180
    RandomAnisotropicScaling:
        params:
            min: 0.8
            max: 1.2
    """
elif data_augmentation == 'GaussianOnly':
    data_augmentation_yaml = """
    RandomGaussianNoise:
        params:
            sigma: 0.01
            clip: 0.05
    """
elif data_augmentation == 'RotationOnly':
    data_augmentation_yaml = """
    RandomRotation:
        params:
            max_degrees: 180
    """
elif data_augmentation == 'ScalingOnly':
    data_augmentation_yaml = """
    RandomAnisotropicScaling:
        params:
            min: 0.8
            max: 1.2
    """
elif data_augmentation == 'No':
    data_augmentation_yaml = """
    """
else:
    print('illegal data augmenatation method! Please use \'Full\', '+\
        '\'GaussianOnly\', \'RotationOnly\', \'ScalingOnly\' or \'No\'!')
    sys.exit()
    
data_augmentation = OmegaConf.create(data_augmentation_yaml)
    
# initialize tensorboard writer
if write_log:
    logdir_path = os.path.join('runs', tensorboard_logdir)
    writer = tensorboard.SummaryWriter(logdir_path)
    
if tree_algorithm == 'simple':
    tree = 'simple'
    max_tree_construction = False
elif tree_algorithm == 'max':
    tree = 'simple'
    max_tree_construction = True
elif tree_algorithm == 'plusplus':
    tree = 'plusplus'
    max_tree_construction = False
elif tree_algorithm == 'plusplusQ':
    tree = 'plusplusQ'
    max_tree_construction = False
else:
    print('illegal tree construction algorithm specified! Please use \'simple\', '+\
        '\'max\', \'plusplus\' or \'plusplusQ\'!')
    sys.exit()
    
if num_trees < 1:
    num_trees = 1
    
if test_loss == 'Accuracy':
    test_loss_fn = Accuracy_Loss()
elif test_loss == 'IoU':
    test_loss_fn = IoU_Loss()
else:
    print('Illegal test loss specified, use'+\
        '\'Accuracy\' or \'IoU\'!')
    sys.exit()
    
if num_heads <= 0:
    num_heads = 1
if dim_hidden <= 0:
    dim_hidden = 1
    
if category not in ShapeNet_Dataloader.seg_classes:
    print('illegal category specified')
    sys.exit()
    
run_performances_file = job_id + '_resultfile.pkl'
time_launched = datetime.now().strftime("%d/%m/%Y %H:%M:%S")
data_to_save = {
    'model batch size': model_batch_size,
    'benchmark batch size': benchmark_batch_size,
    'n steps': n_steps,
    'lr': lr,
    'n training runs': n_training_runs,
    'tensorboard logdir': tensorboard_logdir,
    'write log': write_log,
    'exact timing': exact_timing,
    'ln': ln,
    'tree depth': tree_depth,
    'benchmark': benchmark,
    'model_architecture': model_architecture,
    'warmup steps': warmup_steps,
    'lr decrease interval':lr_decrease_interval,
    'directory': directory,
    'model label': model_label,
    'benchmark label': benchmark_label,
    'plot file': plot_file,
    'data augmentation': data_augmentation,
    'tree algorithm': tree_algorithm,
    'job id': job_id,
    'time launched': time_launched,
    'num trees': num_trees
    }
        
def train_model(model, loss_fn, model_optimizer, num_steps, dataloader,
                warmup_steps, warmup_inc, save_dir=None, verbose=True):
    if save_dir is not None:
        if not os.path.isdir(save_dir):
            os.makedirs(save_dir)

    test_losses = []
    val_losses = []
    steps = []
    
    batches_per_epoch = math.ceil(dataloader.get_train_dataset_size()/model_batch_size)
    print('batches per epoch:')
    print(batches_per_epoch)

    start_time = time()
    
    train_step = 0
    running_loss = 0
    model.train()
    while train_step <= num_steps:
        # fetch random batch
        batch_points, batch_labels = dataloader.get_train_batch(model_batch_size)
        if exact_timing:
            start = time()
        label_probabilities = model(batch_points)
        if exact_timing:
            if device == 'cuda':
                torch.cuda.current_stream().synchronize()
            print('total forward time: ' + str(time() - start))
        # optimize over set transformer parameters
        model_optimizer.zero_grad()
        loss = loss_fn(label_probabilities, batch_labels)
        running_loss += loss.detach()
        if exact_timing:
            start = time()
        loss.backward()
        if exact_timing:
            if device == 'cuda':
                torch.cuda.current_stream().synchronize()
            print('total backward time: ' + str(time() - start))
            print('\n')
        model_optimizer.step()
        
        if write_log and train_step % 50 == 0 and train_step != 0:
            running_loss = running_loss/50
            writer.add_scalar('model loss train', running_loss, train_step)
            running_loss = 0
        
        if train_step < warmup_steps:
            model_optimizer.param_groups[0]['lr'] += warmup_inc
        elif train_step % lr_decrease_interval == 0 and train_step != 0:
            model_optimizer.param_groups[0]['lr'] *= 0.8
            print('lr reducing, new lr: ' + str(model_optimizer.param_groups[0]['lr']))
                
        train_step += 1

        if verbose and train_step % 1000 == 0 and train_step != 0:
            print('-------------------------------')
            print('Step: ' + str(train_step))
            model.eval()
            classification_test_loss = evaluate(model, dataloader, dataset='test')
            classification_val_loss = evaluate(model, dataloader, dataset='val')
            model.train()
            print('Model test loss: ' + str(classification_test_loss))
            print('Model validation loss: ' + str(classification_val_loss))
            test_losses.append(classification_test_loss)
            val_losses.append(classification_val_loss)
            steps.append(train_step)
            if write_log:
                writer.add_scalar('model loss test', classification_test_loss, train_step)
                writer.add_scalar('model loss validation', classification_val_loss,
                                  train_step)
            
    time_elapsed = time() - start_time
    print('Finished model training!')
    print('Total computation time: ' + str(time_elapsed) + 's')

    if save_dir is not None:
        torch.save({'state_dict':model.state_dict()},
            os.path.join(save_dir, 'model.tar'))
    
    return test_losses, val_losses, steps
        
def train_benchmark(model, loss_fn, benchmark_optimizer, num_steps, dataloader,
                    warmup_steps, warmup_inc, save_dir=None, verbose=True):
    if save_dir is not None:
        if not os.path.isdir(save_dir):
            os.makedirs(save_dir)
            
    test_losses = []
    val_losses = []
    steps = []

    batches_per_epoch = math.ceil(dataloader.get_train_dataset_size()/benchmark_batch_size)
    print('batches per epoch:')
    print(batches_per_epoch)

    start_time = time()
    
    train_step = 0
    running_loss = 0
    model.train()
    while train_step <= num_steps:
        benchmark_optimizer.zero_grad()
        processed_batch_size = 2
        while processed_batch_size < benchmark_batch_size:
            # fetch random batch
            batch_points, batch_labels = dataloader.get_train_batch(2)
            label_probabilities = model(batch_points)
            loss = loss_fn(label_probabilities, batch_labels)
            running_loss += loss.detach()*2/benchmark_batch_size
            loss.backward()
            processed_batch_size += 2
        benchmark_optimizer.step()
        
        if write_log and train_step % 50 == 0 and train_step != 0:
            running_loss = running_loss/50
            writer.add_scalar('benchmark loss train', running_loss, train_step)
            running_loss = 0
            
        if train_step < warmup_steps:
            benchmark_optimizer.param_groups[0]['lr'] += warmup_inc
        elif train_step % lr_decrease_interval == 0 and train_step != 0:
            benchmark_optimizer.param_groups[0]['lr'] *= 0.8
            print('lr reducing, new lr: ' + str(benchmark_optimizer.param_groups[0]['lr']))
                
        train_step += 1

        if verbose and train_step % 1000 == 0 and train_step != 0:
            print('-------------------------------')
            print('Step: ' + str(train_step))
            model.eval()
            classification_test_loss = evaluate(model, dataloader, dataset='test')
            classification_val_loss = evaluate(model, dataloader, dataset='val')
            model.train()
            print('Model test loss: ' + str(classification_test_loss))
            print('Model validation loss: ' + str(classification_val_loss))
            test_losses.append(classification_test_loss)
            val_losses.append(classification_val_loss)
            steps.append(train_step)
            if write_log:
                writer.add_scalar('benchmark loss test', classification_test_loss,
                                  train_step)
                writer.add_scalar('benchmark loss validation', classification_val_loss,
                                  train_step)
            
    time_elapsed = time() - start_time
    print('Finished benchmark training!')
    print('Total computation time: ' + str(time_elapsed) + 's')

    if save_dir is not None:
        torch.save({'state_dict':model.state_dict()},
            os.path.join(save_dir, 'model.tar'))
        
    return test_losses, val_losses, steps

def evaluate(model, dataloader, dataset='test'):
    if dataset == 'test':
        points, labels, factors = dataloader.get_test_dataset()
    elif dataset == 'val':
        points, labels, factors = dataloader.get_val_dataset()
    else:
        print('illegal eval dataset specified. Please use \'test\' or \'val\'!')
        return
    
    percentage = torch.tensor([100], dtype=torch.float, device=device)
    
    loss = torch.zeros(1, device=device)
    for i in range(len(points)):
        sample_points = points[i]
        sample_labels = labels[i]
        sample_factor = factors[i]
        
        predicted_labels = torch.argmax(model(sample_points.unsqueeze(0)).squeeze(),
                                        dim=-1)
        # sample_loss = torch.sum(torch.logical_not(predicted_labels == sample_labels))
        # sample_loss = torch.div(sample_loss, sample_labels.shape[0])
        sample_loss = test_loss_fn.compute_loss(predicted_labels, sample_labels)
        sample_loss = torch.mul(sample_factor, sample_loss)
        loss += sample_loss
    
    loss = torch.mul(loss, percentage)
    
    return loss.item()

# dirty solution to circumvent occasionally appearing torch script error
def create_model(num_classes, num_leafs, ln, max_tree_construction, tree, num_trees,
                 num_heads, dim_hidden):
    try:
        if model_architecture == 'Set_Transformer':
            model = Tree_Segmentation_Network(3, num_classes, num_leafs, ln=ln,
                    max_tree_construction=max_tree_construction, tree=tree,
                    num_trees=num_trees, num_heads=num_heads,
                    dim_hidden=dim_hidden).to(device)
        elif model_architecture == 'Enhanced_Set_Transformer':
            model = Enhanced_Tree_Segmentation_Network(3, num_classes, num_leafs, ln=ln,
                    max_tree_construction=max_tree_construction, tree=tree,
                    num_trees=num_trees, num_heads=num_heads,
                    dim_hidden=dim_hidden).to(device)
        elif model_architecture == 'Set_Transformer_Extra_Deep':
            model = Tree_Segmentation_Network(3, num_classes, num_leafs, ln=ln,
                    max_tree_construction=max_tree_construction, tree=tree,
                    num_trees=num_trees, num_heads=num_heads,
                    dim_hidden=dim_hidden, extra_deep=True).to(device)
        elif model_architecture == 'Standard_Transformer':
            model = Standard_Transformer_Segmentation_Network(3, num_classes,
                    num_leafs, ln=ln, max_tree_construction=max_tree_construction,
                    tree=tree, num_trees=num_trees, num_heads=num_heads,
                    dim_hidden=dim_hidden).to(device)
        elif model_architecture == 'Deep_Standard_Transformer':
            model = Deep_Standard_Transformer_Segmentation_Network(3, num_classes,
                    num_leafs, ln=ln, max_tree_construction=max_tree_construction,
                    tree=tree, num_trees=num_trees, num_heads=num_heads,
                    dim_hidden=dim_hidden).to(device)
        elif model_architecture == 'Shallow_Standard_Transformer':
            model = Shallow_Standard_Transformer_Segmentation_Network(3, num_classes,
                    num_leafs, ln=ln, max_tree_construction=max_tree_construction,
                    tree=tree, num_trees=num_trees, num_heads=num_heads,
                    dim_hidden=dim_hidden).to(device)
    except:
        sleep(1)
        print('retrying to create model')
        model = create_model(num_classes, num_leafs, ln, max_tree_construction,
                             tree, num_trees, num_heads, dim_hidden)
    return model

if __name__ == '__main__':
    
    # only for testing on PC
    if device == 'cpu':
        directory = 'tst_datasets'
        tree = 'plusplusQ'
        # model_architecture = 'Standard_Transformer'
        # model_architecture = None
        benchmark = 'IV'
        test_loss_fn = IoU_Loss()
        category='Motorbike'
        # category='Airplane'
    
    dataloader = ShapeNet_Dataloader(directory, category=category,
                                     data_augmentation=data_augmentation)
    min_points, max_points, avg_points, train_dataset_size,\
        validation_dataset_size, test_dataset_size = dataloader.get_point_cloud_statistics()
    print('Dataset Statistics:')
    print('category: ' + category)
    print('min points: ' + str(min_points))
    print('max points: ' + str(max_points))
    print('avg points: ' + str(avg_points))
    print('train dataset size: ' + str(train_dataset_size))
    print('validation dataset size: ' + str(validation_dataset_size))
    print('test dataset size: ' + str(test_dataset_size))
    num_classes = dataloader.get_label_count()
    num_leafs = math.pow(2, tree_depth)
    
    benchmark_performances = []
    model_performances = []
    benchmark_best = 0
    model_best = 0
    
    # estimate running times in seconds of model and benchmark training
    model_training_time = 3600
    benchmark_training_time = 3600
    
    # parameters for warming up lr
    lr_warmup_inc = lr/(warmup_steps+1)
    initial_lr = lr_warmup_inc
        
    for run in range(n_training_runs):
        
        print('\n\nRun number ' + str(run+1))
        
        loss_fn = Classification_Loss()
        
        # create and train model
        if model_architecture != None:
            time_left = 24*3600 - (time() - job_launch_time) - time_margin
            if time_left < model_training_time:
                print('stopping training runs due to no more time left for model training!')
                break
            
            model = create_model(num_classes, num_leafs, ln, max_tree_construction,
                                 tree, num_trees, num_heads, dim_hidden)
            model_optimizer = torch.optim.Adam(model.parameters(), lr=initial_lr)
            
            print('training model')
            model_training_start_time = time()
            model_test_losses, model_val_losses, model_steps =\
                train_model(model, loss_fn, model_optimizer, n_steps, dataloader,
                            warmup_steps, lr_warmup_inc)
            print('model test losses:')
            print(model_test_losses)
            print('model validation losses:')
            print(model_val_losses)
            model_performances.append(model_test_losses)
            model_training_time = time() - model_training_start_time
            
            best_candidate = model_test_losses[np.argmax(model_val_losses)]
            model_best = max(model_best, best_candidate)
        
        # create and train benchmark
        if benchmark != None:
            time_left = 24*3600 - (time() - job_launch_time) - time_margin
            
            if time_left < benchmark_training_time:
                print('stopping training runs due to no more time left for benchmark training!')
                break
            
            if benchmark == 'Full':
                benchmark_model = Full_Segmentation_Network(3, num_classes, ln=ln,
                                num_heads=num_heads, dim_hidden=dim_hidden).to(device)
            elif benchmark == 'Standard_Transformer':
                benchmark_model = Standard_Transformer_notree(3, num_classes, dim_hidden,
                                                              ln=ln).to(device)
            elif 'IV' in benchmark:
                num_inds = benchmark[2:]
                if len(num_inds) != 0:
                    num_inds = int(num_inds)
                else:
                    num_inds = 32
                benchmark_model = IV_Segmentation_Network_deep(3, num_classes, ln=ln,
                                    num_inds=num_inds, num_heads=num_heads,
                                    dim_hidden=dim_hidden).to(device)
            benchmark_optimizer = torch.optim.Adam(benchmark_model.parameters(),
                                                   lr=initial_lr)
            
            print('training benchmark')
            benchmark_training_start_time = time()
            benchmark_test_losses, benchmark_val_losses, benchmark_steps =\
                train_benchmark(benchmark_model, loss_fn, benchmark_optimizer,
                                n_steps, dataloader, warmup_steps, lr_warmup_inc)
            print('benchmark test losses:')
            print(benchmark_test_losses)
            print('benchmark validation losses:')
            print(benchmark_val_losses)
            benchmark_performances.append(benchmark_test_losses)
            benchmark_training_time = time() - benchmark_training_start_time
            
            best_candidate = benchmark_test_losses[np.argmax(benchmark_val_losses)]
            benchmark_best = max(benchmark_best, best_candidate)
            
        print('finished!!')
        
    plot_dict = {}
    
    if model_architecture != None:
        data_to_save.update({
            'single run model performances': model_performances})
        data_to_save.update({'model best': model_best})
        print('model_best: ' + str(model_best))
        
        model_std_devs = np.std(model_performances, axis=0, ddof=1)
        model_performances = np.mean(model_performances, axis=0)
        data_to_save.update({
            'model mean performances': model_performances,
            'model performances std devs': model_std_devs,
            'model epochs': model_steps
            })
        model_plot_list = [model_steps, model_performances, model_std_devs]
        plot_dict.update({model_label: model_plot_list})
        print('Model:')
        print(model_steps)
        print(model_performances)
        print(model_std_devs)
    
    if benchmark != None:
        data_to_save.update({
            'single run benchmark performances': benchmark_performances})
        data_to_save.update({'benchmark best:': benchmark_best})
        print('benchmark best: ' + str(benchmark_best))
        
        benchmark_std_devs = np.std(benchmark_performances, axis=0, ddof=1)
        benchmark_performances = np.mean(benchmark_performances, axis=0)
        data_to_save.update({
            'benchmark mean performances': benchmark_performances,
            'benchmark performances std devs': benchmark_std_devs,
            'benchmark epochs': benchmark_steps
            })
        benchmark_plot_list = [benchmark_steps, benchmark_performances,
                               benchmark_std_devs]
        plot_dict.update({benchmark_label: benchmark_plot_list})
        print('Benchmark:')
        print(benchmark_steps)
        print(benchmark_performances)
        print(benchmark_std_devs)
    
    # save this runs data to file
    path = os.path.join(run_performances_folder, run_performances_file)
    savefile = open(path, "wb")
    pickle.dump(data_to_save, savefile)
    savefile.close()
    
    if plot_dict != {}:
        title = 'Missclassification Error plot'
        x_label = '# training epochs'
        y_label = '% of missclassified points'
        
        mean_average_plot(plot_dict, plot_file, title, x_label, y_label)
    
    
