#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Oct 25 20:10:37 2021

@author: sandro
"""

import torch
import torch.nn as nn
import torch.nn.functional as F
import math
import argparse
import sys

import numpy as np
import os
from time import time
import matplotlib.pyplot as plt
from torch.utils import tensorboard

from MVN_Models import MixtureOfMVNs, MultivariateNormalDiag
from Tree_attention_scaling_simple_tree_no_sparse.tree_attention_models import\
    Tree_Set_Transformer, Full_Set_Transformer, Set_Transformer, Tree_Set_enc_dec,\
        Full_Set_enc_dec

device = "cuda" if torch.cuda.is_available() else "cpu"
print("Using {} device".format(device))

bench_dir = 'bench'  

# parse command line arguments
parser = argparse.ArgumentParser(description='Define min and max amount of points')
parser.add_argument('-min', metavar='N_min', type=np.uint16, default=300,
                    help='lower bound on the amount of points')
parser.add_argument('-max', metavar='N_max', type=np.uint16, default=600,
                    help='upper bound on the amount of points')
parser.add_argument('-batch_size', metavar='batch_size', type=np.uint16, default=10,
                    help='Batch size')
parser.add_argument('-dimension', metavar='dimension', type=np.uint16, default=3,
                    help='Dimension of the points to process')
parser.add_argument('-centers', metavar='cluster_centers', type=np.uint8, default=4,
                    help='Number of cluster centers for MVN distributions')
parser.add_argument('-epochs', metavar='N_epochs', type=np.uint16, default=15000,
                    help='Number of epochs to train the network')
parser.add_argument('-lr', metavar='learning_rate', type=float, default=1e-3,
                    help='Learning rate')
parser.add_argument('-runs', metavar='N_training_runs', type=np.uint8, default=5,
                    help='Number of training repetitions')
parser.add_argument('-logdir', metavar='tensorboard_logdir', type=str, default='',
                    help='Directory where tensorboard logs should be written to')
parser.add_argument('-log', metavar='write_tensorboard_log', type=bool, default=False,
                    help='Wheter to write a tensorboard log or not')
parser.add_argument('-timing', metavar='do_exact_timing', type=bool, default=False,
                    help='Wheter to perform exact timing of fw and bw pass')
parser.add_argument('-ln', metavar='do_layer_normalization', type=bool, default=False,
                    help='Wheter to perform layer normalization in the Set Transformer')
parser.add_argument('-parprint', metavar='print_parameter_max', type=bool, default=False,
                    help='Wheter to print parameter maxima periodically')
parser.add_argument('-tree_depth', metavar='tree_depth', type=int, default=None,
                    help='Depth of the attention tree')
parser.add_argument('-benchmark', metavar='benchmark', type=str, default='Full',
                    help='Wheter to use Full Set Transformer or Basic Set Transformer')
parser.add_argument('-sigma', metavar='sigma', type=float, default=0.6,
                    help='sigma for the gaussian distributions')
parser.add_argument('-save_dir', metavar='save_dir', type=str, default=None,
                    help='directory to save the model and the benchmark')
parser.add_argument('-model', metavar='model', type=str, default='Set_Transformer',
                    help='Wheter to use vanilla transformer or set transformer as model')

args = parser.parse_args()
N_min = args.min
N_max = args.max
B = args.batch_size
D = args.dimension
K = args.centers
max_epochs = args.epochs
lr = args.lr
n_training_runs = args.runs
tensorboard_logdir = args.logdir
write_log = args.log
exact_timing = args.timing
ln = args.ln
print_parameter_max = args.parprint
tree_depth = args.tree_depth
benchmark = args.benchmark
sigma = args.sigma
model_type = args.model
save_dir = args.save_dir

if benchmark not in ['Full', 'Basic']:
    print('illegal benchmark! Please use \'Full\' or \'Basic\'')
    
if N_min >= N_max:
    print('lower and upper bound must be chosen such that lower bound < upper bound!')
    sys.exit()
    
# initialize tensorboard writer
if write_log:
    logdir_path = os.path.join('runs', tensorboard_logdir)
    writer = tensorboard.SummaryWriter(logdir_path)

def generate_bench(N_min, N_max, K, B, D, mog, num_bench, save_file):
    
    if not os.path.isdir(bench_dir):
        os.makedirs(bench_dir)
        
    N_list = np.random.randint(N_min, N_max, num_bench)
    data = []
    ll = 0.
    for N in N_list:
        X, labels, pi, params = mog.sample(B, N, K, return_gt=True)
        ll += mog.log_prob(X, pi, params).item()
        data.append(X)
    bench = [data, ll/num_bench]
    
    if save_file is not None:
        path = os.path.join(bench_dir, save_file)
        torch.save(bench, path)
        
def print_parameter_maximum(model):
    
    max_parameter = 0
    max_name = ''
    for name, param in model.named_parameters():
        if param.requires_grad:
            current_max = torch.max(torch.abs(param))
            if current_max > max_parameter:
                max_parameter = current_max
                max_name = name
    print('max parameter in ' + max_name)
    try:
        max_parameter_print = max_parameter.item()
    except:
        max_parameter_print = max_parameter
    print('max parameter value: ' + str(max_parameter_print))
        
def train_model(model, N_min, N_max, K, B, D, mog, model_optimizer,
                num_epochs, bench_file=None, save_dir=None):
    if save_dir is not None:
        if not os.path.isdir(save_dir):
            os.makedirs(save_dir)

    if bench_file is not None:
        verbose = True
        
        bench_file_path = os.path.join(bench_dir, bench_file)
        bench = torch.load(bench_file_path)
    else:
        verbose = False

    losses = []
    epchs = []

    start_time = time()
    
    for epoch in range(num_epochs+1):
        if epoch % int(0.2*num_epochs) == 0 and epoch != 0:
            model_optimizer.param_groups[0]['lr'] *= 0.6
            print('lr reducing, new lr: ' + str(model_optimizer.param_groups[0]['lr']))
        model.train()
        # genearate random samples
        N = np.random.randint(N_min, N_max)
        X = mog.sample(B, N, K)
        if exact_timing:
            start = time()
        estimation = model(X)
        if exact_timing:
            if device == 'cuda':
                torch.cuda.current_stream().synchronize()
            print('total forward time: ' + str(time() - start))
        # optimize over set transformer parameters
        model_optimizer.zero_grad()
        ll = mog.log_prob(X, *mog.parse(estimation))
        loss = -ll
        if exact_timing:
            start = time()
        loss.backward()
        if exact_timing:
            if device == 'cuda':
                torch.cuda.current_stream().synchronize()
            print('total backward time: ' + str(time() - start))
            print('\n')
        model_optimizer.step()
        
        if write_log and epoch % 50 == 0:
            writer.add_scalar('model loss train', loss, epoch)

        if epoch % 1000 == 0 and verbose and epoch != 0:
            print('-------------------------------')
            print('Epoch: ' + str(epoch))
            loss, oracle_loss = test(bench, mog, model)
            print('Model loss: ' + str(-loss))
            print('Oracle loss: ' + str(-oracle_loss))
            losses.append(-loss)
            epchs.append(epoch)
            if write_log:
                writer.add_scalar('model loss test', -loss, epoch)
        elif epoch % 250 == 0:
            print('--------')
            print('Epoch: ' + str(epoch))
            if print_parameter_max:
                print_parameter_maximum(model)
            
    time_elapsed = time() - start_time
    print('Finished model training!')
    print('Total computation time: ' + str(time_elapsed) + 's')

    if save_dir is not None:
        torch.save({'state_dict':model.state_dict()},
            os.path.join(save_dir, 'model.tar'))
    
    return losses, epchs
        
def train_benchmark(model, N_min, N_max, K, B, D, mog, optimizer, num_epochs,
          bench_file=None, save_dir=None):
    if save_dir is not None:
        if not os.path.isdir(save_dir):
            os.makedirs(save_dir)

    if bench_file is not None:
        verbose = True
        
        bench_file_path = os.path.join(bench_dir, bench_file)
        bench = torch.load(bench_file_path)
    else:
        verbose = False

    start_time = time()
    
    losses = []
    epchs = []
    
    for epoch in range(num_epochs+1):
        if epoch % int(0.2*num_epochs) == 0 and epoch != 0:
            optimizer.param_groups[0]['lr'] *= 0.6
            print('lr reducing, new lr: ' + str(optimizer.param_groups[0]['lr']))
        model.train()
        optimizer.zero_grad()
        N = np.random.randint(N_min, N_max)
        X = mog.sample(B, N, K)
        ll = mog.log_prob(X, *mog.parse(model(X)))
        loss = -ll
        loss.backward()
        optimizer.step()
        
        if write_log and epoch % 50 == 0:
            writer.add_scalar('benchmark loss train', loss, epoch)

        if epoch % 1000 == 0 and verbose and epoch != 0:
            print('-------------------------------')
            print('Epoch: ' + str(epoch))
            loss, oracle_loss = test(bench, mog, model)
            print('Model loss: ' + str(-loss))
            print('Oracle loss: ' + str(-oracle_loss))
            losses.append(-loss)
            epchs.append(epoch)
            if write_log:
                writer.add_scalar('benchmark loss test', -loss, epoch)
        elif epoch % 250 == 0:
            print('--------')
            print('Epoch: ' + str(epoch))
            
    time_elapsed = time() - start_time
    print('Finished benchmark training!')
    print('Total computation time: ' + str(time_elapsed) + 's')

    if save_dir is not None:
        torch.save({'state_dict':model.state_dict()},
            os.path.join(save_dir, 'benchmark.tar'))
        
    return losses, epchs

def test(bench, mog, model):
    model.eval()
    data, oracle_ll = bench
    avg_ll = 0.
    for X in data:
        X = X.to(device)
        avg_ll += mog.log_prob(X, *mog.parse(model(X))).item()
    avg_ll /= len(data)
    
    return avg_ll, oracle_ll

if __name__ == '__main__':
    
    if tree_depth == None:
        depth = math.ceil(math.log(math.sqrt(N_max), 2))
    else:
        depth = tree_depth
        
    num_leafs = math.pow(2, depth)
    num_bench = 500
    
    mvn = MultivariateNormalDiag(D, sigma)
    mog = MixtureOfMVNs(mvn)
    dim_output = 2*D+1
    
    print('N min = ' + str(N_min))
    print('N max = ' + str(N_max))
    bench_file = 'tst_bench_N' + str(N_min) + '-' + str(N_max) + '.pkl'      
    print('generating bench:')
    generate_bench(N_min, N_max, K, B, D, mog, num_bench, bench_file)
    
    bench_file_path = os.path.join(bench_dir, bench_file)
    bench = torch.load(bench_file_path)
    
    Set_Transformer_performances = []
    Tree_Set_Transformer_performances = []
        
    for run in range(n_training_runs):
        
        print('\n\nRun number ' + str(run+1))
        
        if benchmark == 'Full':
            benchmark_model = Full_Set_enc_dec(D, K, dim_output, ln=ln,
                                     dim_hidden=64).to(device)
        elif benchmark == 'Basic':
            benchmark_model = Set_Transformer(D, K, dim_output, ln=ln).to(device)
        print('orig device:')
        print(device)
        if model_type == 'Set_Transformer':
            model = Tree_Set_Transformer(D, K, dim_output, num_leafs, ln=ln, tree='plusplus',
                                         dim_hidden=64).to(device)
        elif model_type == 'Transformer':
            model = Tree_Set_enc_dec(D, K, dim_output, num_leafs, ln=ln,
                                     tree='plusplus', dim_hidden=64).to(device)
        else:
            print('illegal model specified!')
            sys.exit()
        
        benchmark_optimizer = torch.optim.Adam(benchmark_model.parameters(), lr=lr)
        model_optimizer = torch.optim.Adam(model.parameters(), lr=lr)
        
        print('training model')
        model_losses, model_epchs = train_model(model, N_min, N_max, K, B, D, mog,
            model_optimizer, max_epochs, bench_file=bench_file, save_dir=save_dir)
        print(model_losses)
        Tree_Set_Transformer_performances.append(model_losses)
        
        print('training benchmark')
        benchmark_losses, benchmark_epchs = train_benchmark(benchmark_model, N_min,
            N_max, K, B, D, mog, benchmark_optimizer, max_epochs, bench_file=bench_file,
            save_dir=save_dir)
        print(benchmark_losses)
        Set_Transformer_performances.append(benchmark_losses)
            
        print('finished!!')
    
    Set_Transformer_std_devs = np.std(Set_Transformer_performances, axis=0, ddof=1)
    Set_Transformer_performances = np.mean(Set_Transformer_performances, axis=0)
    Tree_Set_Transformer_std_devs = np.std(Tree_Set_Transformer_performances, axis=0,
                                           ddof=1)
    Tree_Set_Transformer_performances = np.mean(Tree_Set_Transformer_performances,
                                                axis=0)
        
    print('Set Transformer:')
    print(Set_Transformer_performances)
    print(Set_Transformer_std_devs)
    print('Simple Tree Set Transformer:')
    print(Tree_Set_Transformer_performances)
    print(Tree_Set_Transformer_std_devs)
    print('Epochs:')
    print(model_epchs)
    print(benchmark_epchs)
    
    fig, ax = plt.subplots(figsize=(14,10))
    ax.set_title('Negative log likelihood plot', fontsize=30)
    ax.set_xlim([np.min(model_epchs),np.max(model_epchs)])
    # ax.set_ylim([0,3])
    ax.tick_params(axis='both', labelsize=15)
    # plot tree isab performance
    lineplot = plt.plot(model_epchs, Tree_Set_Transformer_performances, color='b',
                        label='Simple Tree Set Transformer')
    ax.fill_between(model_epchs, np.subtract(Tree_Set_Transformer_performances,
                                             Tree_Set_Transformer_std_devs),
                    np.add(Tree_Set_Transformer_performances, Tree_Set_Transformer_std_devs),
                    color='b', alpha=0.3)
    # plot benchmark performance
    lineplot = plt.plot(benchmark_epchs, Set_Transformer_performances, color='r',
                        label='"Normal" Set Transformer')
    ax.fill_between(benchmark_epchs, np.subtract(Set_Transformer_performances,
                                                 Set_Transformer_std_devs),
                    np.add(Set_Transformer_performances, Set_Transformer_std_devs),
                    color='r', alpha=0.3)
    ax.legend(fontsize=20)
    plt.savefig('Full_vs_NN_performance_plot_realNN.png')
    plt.show()
    
    
